"""
QR decomposition of ``DNDarray``s.
"""

import collections
import torch
from typing import Tuple

from ..dndarray import DNDarray
from ..manipulations import concatenate
from .. import factories
from .. import communication
from ..types import float32, float64, complex64, complex128

__all__ = ["qr"]


def qr(
    A: DNDarray,
    mode: str = "reduced",
    procs_to_merge: int = 2,
) -> Tuple[DNDarray, DNDarray]:
    r"""
    Calculates the QR decomposition of a 2D ``DNDarray``.
    Factor the matrix ``A`` as *QR*, where ``Q`` is orthonormal and ``R`` is upper-triangular.
    If ``mode = "reduced"``, function returns ``QR(Q=Q, R=R)``, if ``mode = "r"`` function returns ``QR(Q=None, R=R)``

    This function also works for batches of matrices; in this case, the last two dimensions of the input array are considered as the matrix dimensions.
    The output arrays have the same leading batch dimensions as the input array.

    Parameters
    ----------
    A : DNDarray of shape (M, N), of shape (...,M,N) in the batched case
        Array which will be decomposed.
    mode : str, optional
        default "reduced" returns Q and R with dimensions (M, min(M,N)) and (min(M,N), N). Potential batch dimensions are not modified.
        "r" returns only R, with dimensions (min(M,N), N).
    procs_to_merge : int, optional
        This parameter is only relevant for split=0 (-2, in the batched case) and determines the number of processes to be merged at one step during the so-called TS-QR algorithm.
        The default is 2. Higher choices might be faster, but will probably result in higher memory consumption. 0 corresponds to merging all processes at once.
        We only recommend to modify this parameter if you are familiar with the TS-QR algorithm (see the references below).

    Notes
    -----
    The distribution schemes of ``Q`` and ``R`` depend on that of the input ``A``.

        - If ``A`` is distributed along the columns (A.split = 1), so will be ``Q`` and ``R``.

        - If ``A`` is distributed along the rows (A.split = 0), ``Q`` too will have  `split=0`. ``R`` won't be distributed, i.e. `R. split = None`, if ``A`` is tall-skinny, i.e., if
          the largest local chunk of data of ``A`` has at least as many rows as columns. Otherwise, ``R`` will be distributed along the rows as well, i.e., `R.split = 0`.

    Note that the argument `calc_q` allowed in earlier Heat versions is no longer supported; `calc_q = False` is equivalent to `mode = "r"`.
    Unlike ``numpy.linalg.qr()``, `ht.linalg.qr` only supports ``mode="reduced"`` or ``mode="r"`` for the moment, since "complete" may result in heavy memory usage.

    Heats QR function is built on top of PyTorchs QR function, ``torch.linalg.qr()``, using LAPACK (CPU) and MAGMA (CUDA) on
    the backend. Both cases split=0 and split=1 build on a column-block-wise version of stabilized Gram-Schmidt orthogonalization.
    For split=1 (-1, in the batched case), this is directly applied to the local arrays of the input array.
    For split=0, a tall-skinny QR (TS-QR) is implemented for the case of tall-skinny matrices (i.e., the largest local chunk of data has at least as many rows as columns),
    and extended to non tall-skinny matrices by applying a block-wise version of stabilized Gram-Schmidt orthogonalization.

    References
    ----------
    Basic information about QR factorization/decomposition can be found at, e.g.:

        - https://en.wikipedia.org/wiki/QR_factorization,

        - Gene H. Golub and Charles F. Van Loan. 1996. Matrix Computations (3rd Ed.).

    For an extensive overview on TS-QR and its variants we refer to, e.g.,

        - Demmel, James, et al. “Communication-Optimal Parallel and Sequential QR and LU Factorizations.” SIAM Journal on Scientific Computing, vol. 34, no. 1, 2 Feb. 2012, pp. A206–A239., doi:10.1137/080731992.
    """
    if not isinstance(A, DNDarray):
        raise TypeError(f"'A' must be a DNDarray, but is {type(A)}")
    if not isinstance(mode, str):
        raise TypeError(f"'mode' must be a str, but is {type(mode)}")
    if mode not in ["reduced", "r"]:
        if mode == "complete":
            raise NotImplementedError(
                "QR decomposition with 'mode'='complete' is not supported by heat yet. \n Please open an issue on GitHub if you require this feature. \n For now, you can use 'mode'='reduced' or 'r' instead."
            )
        elif mode == "raw":
            raise NotImplementedError(
                "QR decomposition with 'mode'='raw' is neither supported by Heat nor by PyTorch. \n"
            )
        else:
            raise ValueError(f"'mode' must be 'reduced' (default) or 'r', but is {mode}")
    if not isinstance(procs_to_merge, int):
        raise TypeError(f"procs_to_merge must be an int, but is currently {type(procs_to_merge)}")
    if procs_to_merge < 0 or procs_to_merge == 1:
        raise ValueError(
            f"procs_to_merge must be 0 (for merging all processes at once) or at least 2, but is currently {procs_to_merge}"
        )
    if procs_to_merge == 0:
        procs_to_merge = A.comm.size

    if A.dtype not in [float32, float64, complex64, complex128]:
        try:
            torch.linalg.qr(A.larray)
        except (NotImplementedError, RuntimeError) as E:
            raise TypeError(
                f"Array 'A' must have datatype of float32, float64, complex64, or complex128, but has {A.dtype}"
            ) from E
        raise NotImplementedError(
            f"`heat.linalg.qr` is not implemented for dtype {A.dtype}. Please open an issue on GitHub to get this implemented"
        )

    QR = collections.namedtuple("QR", "Q, R")

    if A.ndim == 3:
        single_proc_qr = torch.vmap(torch.linalg.qr, in_dims=0, out_dims=0)
    else:
        single_proc_qr = torch.linalg.qr

    if not A.is_distributed() or A.split < A.ndim - 2:
        # handle the case of a single process or split=None: just PyTorch QR
        Q, R = single_proc_qr(A.larray, mode=mode)

        R = factories.array(R, is_split=A.split, device=A.device)
        if mode == "reduced":
            Q = factories.array(Q, is_split=A.split, device=A.device)
        else:
            Q = None
        return QR(Q, R)

    if A.split == A.ndim - 1:
        # handle the case that A is split along the columns
        # here, we apply a block-wise version of (stabilized) Gram-Schmidt orthogonalization
        # instead of orthogonalizing each column of A individually, we orthogonalize blocks of columns (i.e. the local arrays) at once

        lshapes = A.lshape_map[:, -1]
        lshapes_cum = torch.cumsum(lshapes, 0)
        nprocs = A.comm.size

        if A.shape[-2] >= A.shape[-1]:
            last_row_reached = nprocs
            k = A.shape[-1]
        else:
            last_row_reached = min(torch.argwhere(lshapes_cum >= A.shape[-2]))[0]
            k = A.shape[-2]

        if mode == "reduced":
            Q = factories.zeros(
                A.shape, dtype=A.dtype, split=A.ndim - 1, device=A.device, comm=A.comm
            )

        R = factories.zeros(
            (*A.shape[:-2], k, A.shape[-1]),
            dtype=A.dtype,
            split=A.ndim - 1,
            device=A.device,
            comm=A.comm,
        )
        R_shapes = torch.hstack(
            [
                torch.zeros(1, dtype=torch.int32, device=A.device.torch_device),
                torch.cumsum(R.lshape_map[:, -1], 0),
            ]
        )

        A_columns = A.larray.clone()

        for i in range(last_row_reached + 1):
            # this loop goes through all the column-blocks (i.e. local arrays) of the matrix
            # this corresponds to the loop over all columns in classical Gram-Schmidt

            if i < nprocs - 1:
                k_loc_i = min(A.shape[-2], A.lshape_map[i, -1])
                Q_buf = torch.zeros(
                    (*A.shape[:-1], k_loc_i),
                    dtype=A.larray.dtype,
                    device=A.device.torch_device,
                )

            if A.comm.rank == i:
                # orthogonalize the current block of columns by utilizing PyTorch QR
                Q_curr, R_loc = single_proc_qr(A_columns, mode="reduced")
                if i < nprocs - 1:
                    Q_buf = Q_curr.contiguous()
                if mode == "reduced":
                    Q.larray = Q_curr
                r_size = R.larray[..., R_shapes[i] : R_shapes[i + 1], :].shape[-2]
                R.larray[..., R_shapes[i] : R_shapes[i + 1], :] = R_loc[..., :r_size, :]

            if i < nprocs - 1:
                # broadcast the orthogonalized block of columns to all other processes
                A.comm.Bcast(Q_buf, root=i)

            if A.comm.rank > i:
                # subtract the contribution of the current block of columns from the remaining columns
                R_loc = torch.transpose(torch.conj(Q_buf), -2, -1) @ A_columns
                A_columns -= Q_buf @ R_loc
                r_size = R.larray[..., R_shapes[i] : R_shapes[i + 1], :].shape[-2]
                R.larray[..., R_shapes[i] : R_shapes[i + 1], :] = R_loc[..., :r_size, :]

        if mode == "reduced":
            Q = Q[..., :, :k].balance()
        else:
            Q = None

        return QR(Q, R)

    if A.split == A.ndim - 2:
        # check that data distribution is reasonable for TS-QR
        # we regard a matrix with split = 0 as suitable for TS-QR if its largest local chunk of data has at least as many rows as columns
        biggest_number_of_local_rows = A.lshape_map[:, -2].max().item()
        if biggest_number_of_local_rows < A.shape[-1]:
            column_idx = torch.cumsum(A.lshape_map[:, -2], 0)
            column_idx = column_idx[column_idx < A.shape[-1]]
            column_idx = torch.cat(
                (
                    torch.tensor([0], device=column_idx.device),
                    column_idx,
                    torch.tensor([A.shape[-1]], device=column_idx.device),
                )
            )
            A_copy = A.copy()
            R = A.copy()
            # Block-wise Gram-Schmidt orthogonalization, applied to groups of columns
            offset = 1 if A.shape[-1] <= A.shape[-2] else 2
            for k in range(len(column_idx) - offset):
                # since we only consider a group of columns, TS QR is applied to a tall-skinny matrix
                Qnew, Rnew = qr(
                    A_copy[..., :, column_idx[k] : column_idx[k + 1]],
                    mode="reduced",
                    procs_to_merge=procs_to_merge,
                )

                # usual update of the remaining columns
                if R.comm.rank == k:
                    R.larray[
                        ...,
                        : (column_idx[k + 1] - column_idx[k]),
                        column_idx[k] : column_idx[k + 1],
                    ] = Rnew.larray
                if R.comm.rank > k:
                    R.larray[..., :, column_idx[k] : column_idx[k + 1]] *= 0
                if k < len(column_idx) - 2:
                    coeffs = (
                        torch.transpose(torch.conj(Qnew.larray), -2, -1)
                        @ A_copy.larray[..., :, column_idx[k + 1] :]
                    )
                    R.comm.Allreduce(communication.MPI.IN_PLACE, coeffs)
                    if R.comm.rank == k:
                        R.larray[..., :, column_idx[k + 1] :] = coeffs
                    A_copy.larray[..., :, column_idx[k + 1] :] -= Qnew.larray @ coeffs
                if mode == "reduced":
                    Q = Qnew if k == 0 else concatenate((Q, Qnew), axis=-1)
            if A.shape[-1] < A.shape[-2]:
                R = R[..., : A.shape[-1], :].balance()
            if mode == "reduced":
                return QR(Q, R)
            else:
                return QR(None, R)

        else:
            # in this case the input is tall-skinny and we apply the TS-QR algorithm
            # it follows the implementation of TS-QR for split = 0
            current_procs = [i for i in range(A.comm.size)]
            current_comm = A.comm
            local_comm = current_comm.Split(current_comm.rank // procs_to_merge, A.comm.rank)
            Q_loc, R_loc = single_proc_qr(A.larray, mode=mode)
            R_loc = R_loc.contiguous()
            if mode == "reduced":
                leave_comm = current_comm.Split(current_comm.rank, A.comm.rank)

            level = 1
            while len(current_procs) > 1:
                if A.comm.rank in current_procs and local_comm.size > 1:
                    # create array to collect the R_loc's from all processes of the process group of at most n_procs_to_merge processes
                    shapes_R_loc = local_comm.gather(R_loc.shape[-2], root=0)
                    if local_comm.rank == 0:
                        gathered_R_loc = torch.zeros(
                            (*R_loc.shape[:-2], sum(shapes_R_loc), R_loc.shape[-1]),
                            device=R_loc.device,
                            dtype=R_loc.dtype,
                        )
                        counts = list(shapes_R_loc)
                        displs = torch.cumsum(
                            torch.tensor([0] + shapes_R_loc, dtype=torch.int32), 0
                        ).tolist()[:-1]
                    else:
                        gathered_R_loc = torch.empty(0, device=R_loc.device, dtype=R_loc.dtype)
                        counts = None
                        displs = None
                    # gather the R_loc's from all processes of the process group of at most n_procs_to_merge processes
                    local_comm.Gatherv(R_loc, (gathered_R_loc, counts, displs), root=0, axis=-2)
                    # perform QR decomposition on the concatenated, gathered R_loc's to obtain new R_loc
                    if local_comm.rank == 0:
                        previous_shape = R_loc.shape
                        Q_buf, R_loc = single_proc_qr(gathered_R_loc, mode=mode)
                        R_loc = R_loc.contiguous()
                    else:
                        Q_buf = torch.empty(0, device=R_loc.device, dtype=R_loc.dtype)
                    if mode == "reduced":
                        if local_comm.rank == 0:
                            Q_buf = Q_buf.contiguous()
                        scattered_Q_buf = torch.empty(
                            R_loc.shape if local_comm.rank != 0 else previous_shape,
                            device=R_loc.device,
                            dtype=R_loc.dtype,
                        )
                        # scatter the Q_buf to all processes of the process group
                        local_comm.Scatterv(
                            (Q_buf, counts, displs), scattered_Q_buf, root=0, axis=-2
                        )
                    del gathered_R_loc, Q_buf

                # for each process in the current processes, broadcast the scattered_Q_buf of this process
                # to all leaves (i.e. all original processes that merge to the current process)
                if mode == "reduced" and leave_comm.size > 1:
                    try:
                        scattered_Q_buf_shape = scattered_Q_buf.shape
                    except UnboundLocalError:
                        scattered_Q_buf_shape = None
                    scattered_Q_buf_shape = leave_comm.bcast(scattered_Q_buf_shape, root=0)
                    if scattered_Q_buf_shape is not None:
                        # this is needed to ensure that only those Q_loc get updates that are actually part of the current process group
                        if leave_comm.rank != 0:
                            scattered_Q_buf = torch.empty(
                                scattered_Q_buf_shape, device=Q_loc.device, dtype=Q_loc.dtype
                            )
                        leave_comm.Bcast(scattered_Q_buf, root=0)
                    # update the local Q_loc by multiplying it with the scattered_Q_buf
                try:
                    Q_loc = Q_loc @ scattered_Q_buf
                    del scattered_Q_buf
                except UnboundLocalError:
                    pass

                # update: determine processes to be active at next "merging" level, create new communicator and split it into groups for gathering
                current_procs = [
                    current_procs[i] for i in range(len(current_procs)) if i % procs_to_merge == 0
                ]
                if len(current_procs) > 1:
                    new_group = A.comm.group.Incl(current_procs)
                    current_comm = A.comm.Create_group(new_group)
                    if A.comm.rank in current_procs:
                        local_comm = communication.MPICommunication(
                            current_comm.Split(current_comm.rank // procs_to_merge, A.comm.rank)
                        )
                    if mode == "reduced":
                        leave_comm = A.comm.Split(A.comm.rank // procs_to_merge**level, A.comm.rank)
                level += 1
            # broadcast the final R_loc to all processes
            R_gshape = (*A.shape[:-2], A.shape[-1], A.shape[-1])
            if A.comm.rank != 0:
                R_loc = torch.empty(R_gshape, dtype=R_loc.dtype, device=R_loc.device)
            A.comm.Bcast(R_loc, root=0)
            R = DNDarray(
                R_loc,
                gshape=R_gshape,
                dtype=A.dtype,
                split=None,
                device=A.device,
                comm=A.comm,
                balanced=True,
            )
            if mode == "r":
                Q = None
            else:
                Q = DNDarray(
                    Q_loc,
                    gshape=A.shape,
                    dtype=A.dtype,
                    split=A.split,
                    device=A.device,
                    comm=A.comm,
                    balanced=True,
                )
            return QR(Q, R)



"""
Module implementing the Dynamic Mode Decomposition (DMD) algorithm.
"""

import heat as ht
from typing import Optional, Union, List
import torch

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self


def _torch_matrix_diag(diagonal):
    # auxiliary function to create a batch of diagonal matrices from a batch of diagonal vectors
    # source: fmassas comment on Oct 4, 2018 in https://github.com/pytorch/pytorch/issues/12160 [Accessed Oct 09, 2024]
    N = diagonal.shape[-1]
    shape = diagonal.shape[:-1] + (N, N)
    device, dtype = diagonal.device, diagonal.dtype
    result = torch.zeros(shape, dtype=dtype, device=device)
    indices = torch.arange(result.numel(), device=device).reshape(shape)
    indices = indices.diagonal(dim1=-2, dim2=-1)
    result.view(-1)[indices] = diagonal
    return result


class DMD(ht.RegressionMixin, ht.BaseEstimator):
    """
    Dynamic Mode Decomposition (DMD), plain vanilla version with SVD-based implementation.

    The time series of which DMD shall be computed must be provided as a 2-D DNDarray of shape (n_features, n_timesteps).
    Please, note that this deviates from Heat's convention that data sets are handeled as 2-D arrays with the feature axis being the second axis.

    Parameters
    ----------
    svd_solver : str, optional
        Specifies the algorithm to use for the singular value decomposition (SVD). Options are 'full' (default), 'hierarchical', and 'randomized'.
    svd_rank : int, optional
        The rank to which SVD shall be truncated. For `'full'` SVD, `svd_rank = None` together with `svd_tol = None` (default) will result in no truncation.
        For `svd_solver='full'`, at most one of `svd_rank` or `svd_tol` may be specified.
        For `svd_solver='hierarchical'`, either `svd_rank` (rank to truncate to) or `svd_tol` (tolerance to truncate to) must be specified.
        For `svd_solver='randomized'`, `svd_rank` must be specified and determines the the rank to truncate to.
    svd_tol : float, optional
        The tolerance to which SVD shall be truncated. For `'full'` SVD, `svd_tol = None` together with `svd_rank = None` (default) will result in no truncation.
        For `svd_solver='hierarchical'`, either `svd_tol` (accuracy to truncate to) or `svd_rank` (rank to truncate to) must be specified.
        For `svd_solver='randomized'`, `svd_tol` is meaningless and must be None.

    Attributes
    ----------
    svd_solver : str
        The algorithm used for the singular value decomposition (SVD).
    svd_rank : int
        The rank to which SVD shall be truncated.
    svd_tol : float
        The tolerance to which SVD shall be truncated.
    rom_basis_ : DNDarray
        The reduced order model basis.
    rom_transfer_matrix_ : DNDarray
        The reduced order model transfer matrix.
    rom_eigenvalues_ : DNDarray
        The reduced order model eigenvalues.
    rom_eigenmodes_ : DNDarray
        The reduced order model eigenmodes ("DMD modes")

    Notes
    -----
    We follow the "exact DMD" method as described in [1], Sect. 2.2.

    Please note that "rank" in the context of SVD always refers to the number of singular values/vectors to compute (i.e., "rank" refers to the mathematical rank
    of a matrix). This is completely different from the notion of "(MPI-)rank", i.e., the ID given to a process, in a parallel MPI-application.

    References
    ----------
    [1] J. L. Proctor, S. L. Brunton, and J. N. Kutz, "Dynamic Mode Decomposition with Control," SIAM Journal on Applied Dynamical Systems, vol. 15, no. 1, pp. 142-161, 2016.
    """

    def __init__(
        self,
        svd_solver: Optional[str] = "full",
        svd_rank: Optional[int] = None,
        svd_tol: Optional[float] = None,
    ):
        # Check if the specified SVD algorithm is valid
        if not isinstance(svd_solver, str):
            raise TypeError(
                f"Invalid type '{type(svd_solver)}' for 'svd_solver'. Must be a string."
            )
        # check if the specified SVD algorithm is valid
        if svd_solver not in ["full", "hierarchical", "randomized"]:
            raise ValueError(
                f"Invalid SVD algorithm '{svd_solver}'. Must be one of 'full', 'hierarchical', 'randomized'."
            )
        # check if the respective algorithm got the right combination of non-None parameters
        if svd_solver == "full" and svd_rank is not None and svd_tol is not None:
            raise ValueError(
                "For 'full' SVD, at most one of 'svd_rank' or 'svd_tol' may be specified."
            )
        if svd_solver == "hierarchical":
            if svd_rank is None and svd_tol is None:
                raise ValueError(
                    "For 'hierarchical' SVD, exactly one of 'svd_rank' or 'svd_tol' must be specified, but none of them is specified."
                )
            if svd_rank is not None and svd_tol is not None:
                raise ValueError(
                    "For 'hierarchical' SVD, exactly one of 'svd_rank' or 'svd_tol' must be specified, but currently both are specified."
                )
        if svd_solver == "randomized":
            if svd_rank is None:
                raise ValueError("For 'randomized' SVD, 'svd_rank' must be specified.")
            if svd_tol is not None:
                raise ValueError("For 'randomized' SVD, 'svd_tol' must be None.")
        # check correct data types of non-None parameters
        if svd_rank is not None:
            if not isinstance(svd_rank, int):
                raise TypeError(
                    f"Invalid type '{type(svd_rank)}' for 'svd_rank'. Must be an integer."
                )
            if svd_rank < 1:
                raise ValueError(
                    f"Invalid value '{svd_rank}' for 'svd_rank'. Must be a positive integer."
                )
        if svd_tol is not None:
            if not isinstance(svd_tol, float):
                raise TypeError(f"Invalid type '{type(svd_tol)}' for 'svd_tol'. Must be a float.")
            if svd_tol <= 0:
                raise ValueError(f"Invalid value '{svd_tol}' for 'svd_tol'. Must be non-negative.")
        # set or initialize the attributes
        self.svd_solver = svd_solver
        self.svd_rank = svd_rank
        self.svd_tol = svd_tol
        self.rom_basis_ = None
        self.rom_transfer_matrix_ = None
        self.rom_eigenvalues_ = None
        self.rom_eigenmodes_ = None
        self.dmdmodes_ = None
        self.n_modes_ = None

    def fit(self, X: ht.DNDarray) -> Self:
        """
        Fits the DMD model to the given data.

        Parameters
        ----------
        X : DNDarray
            The time series data to fit the DMD model to. Must be of shape (n_features, n_timesteps).
        """
        ht.sanitize_in(X)
        # check if the input data is a 2-D DNDarray
        if X.ndim != 2:
            raise ValueError(
                f"Invalid shape '{X.shape}' for input data 'X'. Must be a 2-D DNDarray of shape (n_features, n_timesteps)."
            )
        # check if the input data has at least two time steps
        if X.shape[1] < 2:
            raise ValueError(
                f"Invalid number of time steps '{X.shape[1]}' in input data 'X'. Must have at least two time steps."
            )
        # first step of DMD: compute the SVD of the input data from first to second last time step
        if self.svd_solver == "full" or not X.is_distributed():
            U, S, Vt = ht.linalg.svd(
                X[:, :-1] if X.split == 0 else X[:, :-1].balance(), full_matrices=False
            )
            if self.svd_tol is not None:
                # truncation w.r.t. prescribed bound on explained variance
                # determine svd_rank accordingly
                total_variance = (S**2).sum()
                variance_threshold = (1 - self.svd_tol) * total_variance.larray.item()
                variance_cumsum = (S**2).larray.cumsum(0)
                self.n_modes_ = len(variance_cumsum[variance_cumsum <= variance_threshold]) + 1
            elif self.svd_rank is not None:
                # truncation w.r.t. prescribed rank
                self.n_modes_ = self.svd_rank
            else:
                # no truncation
                self.n_modes_ = S.shape[0]
            self.rom_basis_ = U[:, : self.n_modes_]
            Vt = Vt[: self.n_modes_, :]
            S = S[: self.n_modes_]
        # compute SVD via "hierarchical" SVD
        elif self.svd_solver == "hierarchical":
            if self.svd_tol is not None:
                # hierarchical SVD with prescribed upper bound on relative error
                U, S, Vt, _ = ht.linalg.hsvd_rtol(
                    X[:, :-1] if X.split == 0 else X[:, :-1].balance(),
                    self.svd_tol,
                    compute_sv=True,
                    safetyshift=5,
                )
            else:
                # hierarchical SVD with prescribed, fixed rank
                U, S, Vt, _ = ht.linalg.hsvd_rank(
                    X[:, :-1] if X.split == 0 else X[:, :-1].balance(),
                    self.svd_rank,
                    compute_sv=True,
                    safetyshift=5,
                )
            self.rom_basis_ = U
            self.n_modes_ = U.shape[1]
        else:
            # compute SVD via "randomized" SVD
            U, S, Vt = ht.linalg.rsvd(
                X[:, :-1] if X.split == 0 else X[:, :-1].balance_(),
                self.svd_rank,
            )
            self.rom_basis_ = U
            self.n_modes_ = U.shape[1]
        # second step of DMD: compute the reduced order model transfer matrix
        # we need to assume that the transfer matrix of the ROM is small enough to fit into memory of one process
        if X.split == 0 or X.split is None:
            # if split axis of the input data is 0, using X[:,1:] does not result in un-balancedness and corresponding problems in matmul
            self.rom_transfer_matrix_ = self.rom_basis_.T @ X[:, 1:] @ Vt.T / S
        else:
            # if input is split along columns, X[:,1:] will be un-balanced and cause problems in matmul
            Xplus = X[:, 1:]
            Xplus.balance_()
            self.rom_transfer_matrix_ = self.rom_basis_.T @ Xplus @ Vt.T / S

        self.rom_transfer_matrix_.resplit_(None)
        # third step of DMD: compute the reduced order model eigenvalues and eigenmodes
        eigvals_loc, eigvec_loc = torch.linalg.eig(self.rom_transfer_matrix_.larray)
        self.rom_eigenvalues_ = ht.array(eigvals_loc, split=None, device=X.device)
        self.rom_eigenmodes_ = ht.array(eigvec_loc, split=None, device=X.device)
        self.dmdmodes_ = self.rom_basis_ @ self.rom_eigenmodes_

    def predict_next(self, X: ht.DNDarray, n_steps: int = 1) -> ht.DNDarray:
        """
        Predicts and returns the state(s) after n_steps-many time steps for given a current state(s).

        Parameters
        ----------
        X : DNDarray
            The current state(s) for the prediction. Must have the same number of features as the training data, but can be batched for multiple current states,
            i.e., X can be of shape (n_features,) or (n_features, n_current_states).
            The output will have the same shape as the input.
        n_steps : int, optional
            The number of steps to predict into the future. Default is 1, i.e., the next time step is predicted.
        """
        if not isinstance(n_steps, int):
            raise TypeError(f"Invalid type '{type(n_steps)}' for 'n_steps'. Must be an integer.")
        if self.rom_basis_ is None:
            raise RuntimeError("Model has not been fitted yet. Call 'fit' first.")
        # sanitize input data
        ht.sanitize_in(X)
        # if X is a 1-D DNDarray, we add an artificial batch dimension
        if X.ndim == 1:
            X = X.expand_dims(1)
        # check if the input data has the right number of features
        if X.shape[0] != self.rom_basis_.shape[0]:
            raise ValueError(
                f"Invalid number of features '{X.shape[0]}' in input data 'X'. Must have the same number of features as the training data."
            )
        rom_mat = self.rom_transfer_matrix_.copy()
        rom_mat.larray = torch.linalg.matrix_power(rom_mat.larray, n_steps)
        # the following line looks that complicated because we have to make sure that splits of the resulting matrices in
        # each of the products are split along the axis that deserves being splitted
        nextX = (self.rom_basis_.T @ X).T.resplit_(None) @ (self.rom_basis_ @ rom_mat).T
        return (nextX.T).squeeze()

    def predict(self, X: ht.DNDarray, steps: Union[int, List[int]]) -> ht.DNDarray:
        """
        Predics and returns future states given a current state(s) and returns them all as an array of size (n_steps, n_features).

        This function avoids a time-stepping loop (i.e., repeated calls to 'predict_next') and computes the future states in one go.
        To do so, the number of future times to predict must be of moderate size as an array of shape (n_steps, self.n_modes_, self.n_modes_) must fit into memory.
        Moreover, it must be ensured that:

        - the array of initial states is not split or split along the batch axis (axis 1) and the feature axis is small (i.e., self.rom_basis_ is not split)

        Parameters
        ----------
        X : DNDarray
            The current state(s) for the prediction. Must have the same number of features as the training data, but can be batched for multiple current states,
            i.e., X can be of shape (n_features,) or (n_current_states, n_features).
        steps : int or List[int]
            if int: predictions at time step 0, 1, ..., steps-1 are computed
            if List[int]: predictions at time steps given in the list are computed
        """
        if self.rom_basis_ is None:
            raise RuntimeError("Model has not been fitted yet. Call 'fit' first.")
        # sanitize input data
        ht.sanitize_in(X)
        # if X is a 1-D DNDarray, we add an artificial batch dimension
        if X.ndim == 1:
            X = X.expand_dims(1)
        # check if the input data has the right number of features
        if X.shape[0] != self.rom_basis_.shape[0]:
            raise ValueError(
                f"Invalid number of features '{X.shape[0]}' in input data 'X'. Must have the same number of features as the training data."
            )
        if isinstance(steps, int):
            steps = torch.arange(steps, dtype=torch.int32, device=X.device.torch_device)
        elif isinstance(steps, list):
            steps = torch.tensor(steps, dtype=torch.int32, device=X.device.torch_device)
        else:
            raise TypeError(
                f"Invalid type '{type(steps)}' for 'steps'. Must be an integer or a list of integers."
            )
        steps = steps.reshape(-1, 1).repeat(1, self.rom_eigenvalues_.shape[0])
        X_rom = self.rom_basis_.T @ X

        transfer_mat = _torch_matrix_diag(torch.pow(self.rom_eigenvalues_.larray, steps))
        transfer_mat = (
            self.rom_eigenmodes_.larray @ transfer_mat @ self.rom_eigenmodes_.larray.inverse()
        )
        transfer_mat = torch.real(
            transfer_mat
        )  # necessary to avoid imaginary parts due to numerical errors

        if self.rom_basis_.split is None and (X.split is None or X.split == 1):
            result = (
                transfer_mat @ X_rom.larray
            )  # here we assume that X_rom is not split or split along the second axis (axis 1)
            del transfer_mat

            result = (
                self.rom_basis_.larray @ result
            )  # here we assume that self.rom_basis_ is not split (i.e., the feature number is small)
            result = ht.array(result, is_split=2 if X.split == 1 else None)
            return result.squeeze().T
        else:
            raise NotImplementedError(
                "Predicting multiple time steps in one go is not supported for the given data layout. Please, use 'predict_next' instead, or open an issue on GitHub if you require this feature."
            )

    def __str__(self):
        if ht.MPI_WORLD.rank == 0:
            if self.rom_basis_ is not None:
                return (
                    f"-------------------- DMD (Dynamic Mode Decomposition) --------------------\n"
                    f"Number of modes: {self.n_modes_}\n"
                    f"State space dimension: {self.rom_basis_.shape[0]}\n"
                    f"DMD eigenvalues: {self.rom_eigenvalues_.larray}\n"
                    f"--------------------------------------------------------------------------\n"
                    f"ROM basis of shape {self.rom_basis_.shape}:\n"
                    f"\t split axis: {self.rom_basis_.split}\n"
                    f"\t device: {self.rom_basis_.device.__str__().split(':')[-2]}\n"
                    f"--------------------------------------------------------------------------\n"
                )
            else:
                return (
                    f"---------------- UNFITTED DMD (Dynamic Mode Decomposition) ---------------\n"
                    f"Parameters for fit are as follows: \n"
                    f"\t SVD-solver: {self.svd_solver}\n"
                    f"\t SVD-rank: {self.svd_rank}\n"
                    f"\t SVD-tolerance: {self.svd_tol}\n"
                    f"--------------------------------------------------------------------------\n"
                )
        else:
            return ""


class DMDc(ht.RegressionMixin, ht.BaseEstimator):
    """
    Dynamic Mode Decomposition with Control (DMDc), plain vanilla version with SVD-based implementation.

    The time series of states and controls must be provided as 2-D DNDarrays of shapes (n_state_features, n_timesteps) and (n_control_features, n_timesteps), respectively.
    Please, note that this deviates from Heat's convention that data sets are handeled as 2-D arrays with the feature axis being the second axis.

    Parameters
    ----------
    svd_solver : str, optional
        Specifies the algorithm to use for the singular value decomposition (SVD). Options are 'full' (default), 'hierarchical', and 'randomized'.
    svd_rank : int, optional
        The rank to which SVD of the states shall be truncated. For `'full'` SVD, `svd_rank = None` together with `svd_tol = None` (default) will result in no truncation.
        For `svd_solver='full'`, at most one of `svd_rank` or `svd_tol` may be specified.
        For `svd_solver='hierarchical'`, either `svd_rank` (rank to truncate to) or `svd_tol` (tolerance to truncate to) must be specified.
        For `svd_solver='randomized'`, `svd_rank` must be specified and determines the the rank to truncate to.
    svd_tol : float, optional
        The tolerance to which SVD of the states shall be truncated. For `'full'` SVD, `svd_tol = None` together with `svd_rank = None` (default) will result in no truncation.
        For `svd_solver='hierarchical'`, either `svd_tol` (accuracy to truncate to) or `svd_rank` (rank to truncate to) must be specified.
        For `svd_solver='randomized'`, `svd_tol` is meaningless and must be None.

    Attributes
    ----------
    svd_solver : str
        The algorithm used for the singular value decomposition (SVD).
    svd_rank : int
        The rank to which SVD shall be truncated.
    svd_tol : float
        The tolerance to which SVD shall be truncated.
    rom_basis_ : DNDarray
        The reduced order model basis.
    rom_transfer_matrix_ : DNDarray
        The reduced order model transfer matrix.
    rom_control_matrix_ : DNDarray
        The reduced order model control matrix.
    rom_eigenvalues_ : DNDarray
        The reduced order model eigenvalues.
    rom_eigenmodes_ : DNDarray
        The reduced order model eigenmodes ("DMD modes")

    Notes
    -----
    We follow the approach described in [1], Sects. 3.3 and 3.4.
    In the case that svd_rank is prescribed, the rank of the SVD of the full system matrix is set to svd_rank + n_control_features; cf. https://github.com/dynamicslab/pykoopman
    for the same approach.

    Please note that "rank" in the context of SVD always refers to the number of singular values/vectors to compute (i.e., "rank" refers to the mathematical rank
    of a matrix). This is completely different from the notion of "(MPI-)rank", i.e., the ID given to a process, in a parallel MPI-application.

    References
    ----------
    [1] J. L. Proctor, S. L. Brunton, and J. N. Kutz, "Dynamic Mode Decomposition with Control," SIAM Journal on Applied Dynamical Systems, vol. 15, no. 1, pp. 142-161, 2016.
    """

    def __init__(
        self,
        svd_solver: Optional[str] = "full",
        svd_rank: Optional[int] = None,
        svd_tol: Optional[float] = None,
    ):
        # Check if the specified SVD algorithm is valid
        if not isinstance(svd_solver, str):
            raise TypeError(
                f"Invalid type '{type(svd_solver)}' for 'svd_solver'. Must be a string."
            )
        # check if the specified SVD algorithm is valid
        if svd_solver not in ["full", "hierarchical", "randomized"]:
            raise ValueError(
                f"Invalid SVD algorithm '{svd_solver}'. Must be one of 'full', 'hierarchical', 'randomized'."
            )
        # check if the respective algorithm got the right combination of non-None parameters
        if svd_solver == "full" and svd_rank is not None and svd_tol is not None:
            raise ValueError(
                "For 'full' SVD, at most one of 'svd_rank' or 'svd_tol' may be specified."
            )
        if svd_solver == "hierarchical":
            if svd_rank is None and svd_tol is None:
                raise ValueError(
                    "For 'hierarchical' SVD, exactly one of 'svd_rank' or 'svd_tol' must be specified, but none of them is specified."
                )
            if svd_rank is not None and svd_tol is not None:
                raise ValueError(
                    "For 'hierarchical' SVD, exactly one of 'svd_rank' or 'svd_tol' must be specified, but currently both are specified."
                )
        if svd_solver == "randomized":
            if svd_rank is None:
                raise ValueError("For 'randomized' SVD, 'svd_rank' must be specified.")
            if svd_tol is not None:
                raise ValueError("For 'randomized' SVD, 'svd_tol' must be None.")
        # check correct data types of non-None parameters
        if svd_rank is not None:
            if not isinstance(svd_rank, int):
                raise TypeError(
                    f"Invalid type '{type(svd_rank)}' for 'svd_rank'. Must be an integer."
                )
            if svd_rank < 1:
                raise ValueError(
                    f"Invalid value '{svd_rank}' for 'svd_rank'. Must be a positive integer."
                )
        if svd_tol is not None:
            if not isinstance(svd_tol, float):
                raise TypeError(f"Invalid type '{type(svd_tol)}' for 'svd_tol'. Must be a float.")
            if svd_tol <= 0:
                raise ValueError(f"Invalid value '{svd_tol}' for 'svd_tol'. Must be non-negative.")
        # set or initialize the attributes
        self.svd_solver = svd_solver
        self.svd_rank = svd_rank
        self.svd_tol = svd_tol
        self.rom_basis_ = None
        self.rom_transfer_matrix_ = None
        self.rom_control_matrix_ = None
        self.rom_eigenvalues_ = None
        self.rom_eigenmodes_ = None
        self.dmdmodes_ = None
        self.n_modes_ = None
        self.n_modes_system_ = None

    def fit(self, X: ht.DNDarray, C: ht.DNDarray) -> Self:
        """
        Fits the DMD model to the given data.

        Parameters
        ----------
        X : DNDarray
            The time series data of states to fit the DMD model to. Must be of shape (n_state_features, n_timesteps).
        C : DNDarray
            The time series of control inputs to fit the DMD model to. Must be of shape (n_control_features, n_timesteps).
        """
        ht.sanitize_in(X)
        ht.sanitize_in(C)
        # check if the input data is a 2-D DNDarray
        if X.ndim != 2:
            raise ValueError(
                f"Invalid shape '{X.shape}' for input data 'X'. Must be a 2-D DNDarray of shape (n_state_features, n_timesteps)."
            )
        if C.ndim != 2:
            raise ValueError(
                f"Invalid shape '{C.shape}' for input data 'C'. Must be a 2-D DNDarray of shape (n_control_features, n_timesteps)."
            )
        # check if the input data has at least two time steps
        if X.shape[1] < 2:
            raise ValueError(
                f"Invalid number of time steps '{X.shape[1]}' in input data 'X'. Must have at least two time steps."
            )
        if C.shape[1] < 2:
            raise ValueError(
                f"Invalid number of time steps '{C.shape[1]}' in input data 'C'. Must have at least two time steps."
            )
        # check if the input data has the same number of time steps
        if X.shape[1] != C.shape[1]:
            raise ValueError(
                f"Invalid number of time steps {X.shape[1]} in input data 'X' and {C.shape[1]} in input data 'C'. Must have the same number of time steps."
            )
        if X.split is not None and C.split is not None and X.split != C.split:
            raise ValueError(
                f"If both input data 'X' and 'C' are distributed, they must be distributed along the same axis, but X.split={X.split}, C.split={C.split}."
            )
        Xplus = X[:, 1:]
        Xplus.balance_()
        Omega = ht.concatenate((X, C), axis=0)[:, :-1]
        # first step of DMDc: compute the SVD of the input data from first to second last time step
        # as well as of the full system matrix
        if self.svd_solver == "full" or not X.is_distributed():
            U, S, Vt = ht.linalg.svd(Xplus, full_matrices=False)
            Utilde, Stilde, Vttilde = ht.linalg.svd(Omega, full_matrices=False)
            if self.svd_tol is not None:
                # truncation w.r.t. prescribed bound on explained variance
                # determine svd_rank accordingly
                total_variance = (S**2).sum()
                variance_threshold = (1 - self.svd_tol) * total_variance.larray.item()
                variance_cumsum = (S**2).larray.cumsum(0)
                self.n_modes_ = len(variance_cumsum[variance_cumsum <= variance_threshold]) + 1
                total_variance_system = (Stilde**2).sum()
                variance_threshold_system = (1 - self.svd_tol) * total_variance_system.larray.item()
                variance_cumsum_system = (Stilde**2).larray.cumsum(0)
                self.n_modes_system_ = (
                    len(variance_cumsum_system[variance_cumsum_system <= variance_threshold_system])
                    + 1
                )
            elif self.svd_rank is not None:
                # truncation w.r.t. prescribed rank
                self.n_modes_ = self.svd_rank
                self.n_modes_system_ = self.svd_rank + C.shape[0]
            else:
                # no truncation
                self.n_modes_ = S.shape[0]
                self.n_modes_system = Stilde.shape[0]

            self.rom_basis_ = U[:, : self.n_modes_]
            Vt = Vt[: self.n_modes_, :]
            S = S[: self.n_modes_]
            Vttilde = Vttilde[: self.n_modes_system_, :]
            Stilde = Stilde[: self.n_modes_system_]
            Utilde1 = Utilde[: X.shape[0], : self.n_modes_system_]
            Utilde2 = Utilde[X.shape[0] :, : self.n_modes_system_]
        # compute SVD via "hierarchical" SVD
        elif self.svd_solver == "hierarchical":
            if self.svd_tol is not None:
                # hierarchical SVD with prescribed upper bound on relative error
                U, S, Vt, _ = ht.linalg.hsvd_rtol(
                    Xplus,
                    self.svd_tol,
                    compute_sv=True,
                    safetyshift=5,
                )
                Utilde, Stilde, Vttilde, _ = ht.linalg.hsvd_rtol(
                    Omega,
                    self.svd_tol,
                    compute_sv=True,
                    safetyshift=5,
                )
            else:
                # hierarchical SVD with prescribed, fixed rank
                U, S, Vt, _ = ht.linalg.hsvd_rank(
                    Xplus,
                    self.svd_rank,
                    compute_sv=True,
                    safetyshift=5,
                )
                Utilde, Stilde, Vttilde, _ = ht.linalg.hsvd_rank(
                    Omega,
                    self.svd_rank + C.shape[0],
                    compute_sv=True,
                    safetyshift=5,
                )
            self.rom_basis_ = U
            self.n_modes_ = U.shape[1]
            self.n_modes_system_ = Utilde.shape[1]
            Utilde1 = Utilde[: X.shape[0], :]
            Utilde2 = Utilde[X.shape[0] :, :]
        else:
            # compute SVD via "randomized" SVD
            U, S, Vt = ht.linalg.rsvd(
                Xplus,
                self.svd_rank,
            )
            Utilde, Stilde, Vttilde = ht.linalg.rsvd(
                Omega,
                self.svd_rank + C.shape[0],
            )
            self.rom_basis_ = U
            self.n_modes_ = U.shape[1]
            self.n_modes_system_ = Utilde.shape[1]
            Utilde1 = Utilde[: X.shape[0], :]
            Utilde2 = Utilde[X.shape[0] :, :]

        # ensure that everything is balanced for the following steps
        Utilde2.balance_()
        Utilde1.balance_()
        Vtilde = Vttilde.T
        Vtilde.balance_()
        if Utilde2.split is not None and Utilde2.shape[Utilde2.split] < Utilde2.comm.size:
            Utilde2.resplit_((Utilde2.split + 1) % 2)
        if Utilde1.split is not None and Utilde1.shape[Utilde1.split] < Utilde1.comm.size:
            Utilde1.resplit_((Utilde1.split + 1) % 2)
        if Vtilde.split is not None and Vtilde.shape[Vtilde.split] < Vtilde.comm.size:
            Vtilde.resplit_((Vtilde.split + 1) % 2)
        # second step of DMD: compute the reduced order model transfer matrix
        # we need to assume that the the transfer matrix of the ROM is small enough to fit into memory of one process
        self.rom_transfer_matrix_ = (
            self.rom_basis_.T
            @ Xplus
            @ (Vtilde / Stilde)
            @ (Utilde1.T @ self.rom_basis_).resplit_(None)
        )
        self.rom_control_matrix_ = (self.rom_basis_.T @ Xplus) @ (
            (Vtilde / Stilde) @ Utilde2.T
        ).resplit_(0)
        self.rom_transfer_matrix_.resplit_(None)
        self.rom_control_matrix_.resplit_(None)

        # third step of DMD: compute the reduced order model eigenvalues and eigenmodes
        eigvals_loc, eigvec_loc = torch.linalg.eig(self.rom_transfer_matrix_.larray)
        self.rom_eigenvalues_ = ht.array(eigvals_loc, split=None, device=X.device)
        self.rom_eigenmodes_ = ht.array(eigvec_loc, split=None, device=X.device)
        self.dmdmodes_ = (
            Xplus @ (Vtilde / Stilde) @ Utilde1.T @ self.rom_basis_ @ self.rom_eigenmodes_
        )

    def predict(self, X: ht.DNDarray, C: ht.DNDarray) -> ht.DNDarray:
        """
        Predicts and returns future states given the current state(s) ``X`` and control trajectory ``C``.

        Parameters
        ----------
        X : DNDarray
            The current state(s) for the prediction. Must have the same number of features as the training data, but can be batched for multiple current states,
            i.e., X can be of shape (n_state_features,) or (n_batch, n_state_features).
        C : DNDarray
            The control trajectory for the prediction. Must have the same number of control features as the training data, i.e., C must be of shape
            (n_control_features,) --for a single time step-- or (n_control_features, n_timesteps).
        """
        if self.rom_basis_ is None:
            raise RuntimeError("Model has not been fitted yet. Call 'fit' first.")
        # sanitize input data
        ht.sanitize_in(X)
        ht.sanitize_in(C)
        # if X is a 1-D DNDarray, we add an artificial batch dimension; check correct dimensions for X
        if X.ndim == 1:
            X = X.expand_dims(0)
        if X.ndim > 2:
            raise ValueError(
                f"Invalid shape '{X.shape}' for input data 'X'. Must be a 2-D DNDarray of shape (n_batch, n_state_features) or a 1-D DNDarray of shape (n_state_features,)."
            )
        # if C is a 1-D DNDarray, we add an artificial dimension for the single time step; check correct dimensions for C
        if C.ndim == 1:
            C = C.expand_dims(1)
        if C.ndim > 2:
            raise ValueError(
                f"Invalid shape '{C.shape}' for input data 'C'. Must be a 2-D DNDarray of shape (n_control_features, n_timesteps) or a 1-D DNDarray of shape (n_control_features,) for a single time step."
            )
        # check if the input data has the right number of features for control and state space
        if X.shape[1] != self.rom_basis_.shape[0]:
            raise ValueError(
                f"Invalid number of features '{X.shape[1]}' in input data 'X'. Must have the same number of features as the training data (={self.rom_basis_.shape[0]})."
            )
        if C.shape[0] != self.rom_control_matrix_.shape[1]:
            raise ValueError(
                f"Invalid number of features '{C.shape[1]}' in input data 'C'. Must have the same number of features as the training data (={self.rom_control_matrix_.shape[1]})."
            )
        # different cases
        if C.split is not None:
            raise ValueError("So far only C.split = None is supported .")
        # time evolution in the reduced order model
        X_red = X @ self.rom_basis_
        X_red_full = ht.zeros(
            (X.shape[0], self.rom_basis_.shape[1], C.shape[1]),
            split=X_red.split,
            device=X.device,
            dtype=X.dtype,
        )
        X_red_full[:, :, 0] = X_red
        for i in range(1, C.shape[1]):
            X_red_full[:, :, i] = (self.rom_transfer_matrix_ @ X_red_full[:, :, i - 1].T).T + (
                self.rom_control_matrix_ @ C[:, i - 1]
            ).T
        # reshape in order to be able to multiply with basis again
        X_red_full = X_red_full.reshape(self.rom_basis_.shape[1], -1).resplit_(
            1 if X_red_full.split == 0 else None
        )
        X_pred = self.rom_basis_ @ X_red_full
        # reshape again and return
        return X_pred.reshape(X.shape[0], X.shape[1], C.shape[1])

    def __str__(self):
        if ht.MPI_WORLD.rank == 0:
            if self.rom_basis_ is not None:
                return (
                    f"----------- DMDc (Dynamic Mode Decomposition with control) ---------------\n"
                    f"Number of modes: {self.n_modes_}\n"
                    f"State space dimension: {self.rom_basis_.shape[0]}\n"
                    f"Control space dimension: {self.rom_control_matrix_.shape[1]}\n"
                    f"DMD eigenvalues: {self.rom_eigenvalues_.larray}\n"
                    f"--------------------------------------------------------------------------\n"
                    f"ROM basis of shape {self.rom_basis_.shape}:\n"
                    f"\t split axis: {self.rom_basis_.split}\n"
                    f"\t device: {self.rom_basis_.device.__str__().split(':')[-2]}\n"
                    f"--------------------------------------------------------------------------\n"
                )
            else:
                return (
                    f"-------- UNFITTED DMDc (Dynamic Mode Decomposition with control) ---------\n"
                    f"Parameters for fit are as follows: \n"
                    f"\t SVD-solver: {self.svd_solver}\n"
                    f"\t SVD-rank: {self.svd_rank}\n"
                    f"\t SVD-tolerance: {self.svd_tol}\n"
                    f"--------------------------------------------------------------------------\n"
                )
        else:
            return ""



from __future__ import annotations

from ._array_object import Array

from typing import NamedTuple
from ._typing import int64

import heat as ht


class UniqueInverseResult(NamedTuple):
    """
    Internal object for return type of ``unique_inverse``.
    """

    values: Array
    inverse_indices: Array


def unique_inverse(x: Array, /) -> UniqueInverseResult:
    """
    Returns the unique elements of an input array ``x`` and the indices from the
    set of unique elements that reconstruct ``x``.

    Parameters
    ----------
    x : Array
        Input array. If ``x`` has more than one dimension, the function flattens ``x``
        and returns the unique elements of the flattened array.
    """
    values, inverse_indices = ht.unique(x._array, return_inverse=True)
    inverse_indices = inverse_indices.astype(int64)
    return UniqueInverseResult(Array._new(values), Array._new(inverse_indices))


def unique_values(x: Array, /) -> Array:
    """
    Returns the unique elements of an input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. If ``x`` has more than one dimension, the function flattens ``x``
        and returns the unique elements of the flattened array.
    """
    return Array._new(ht.unique(x._array))



from __future__ import annotations

import math
import numpy as np
import torch
import warnings

from inspect import stack
from mpi4py import MPI
from pathlib import Path
from typing import List, Union, Tuple, TypeVar, Optional, Iterable
from heat.core.dndarray import DNDarray

def __getitem__(self, key: Union[int, Tuple[int, ...], List[int, ...]]) -> DNDarray:
    """
    Global getter function for DNDarrays.
    Returns a new DNDarray composed of the elements of the original tensor selected by the indices
    given. This does *NOT* redistribute or rebalance the resulting tensor. If the selection of values is
    unbalanced then the resultant tensor is also unbalanced!
    To redistributed the ``DNDarray`` use :func:`balance()` (issue #187)

    Parameters
    ----------
    key : int, slice, Tuple[int,...], List[int,...]
        Indices to get from the tensor.

    Examples
    --------
    >>> a = ht.arange(10, split=0)
    (1/2) >>> tensor([0, 1, 2, 3, 4], dtype=torch.int32)
    (2/2) >>> tensor([5, 6, 7, 8, 9], dtype=torch.int32)
    >>> a[1:6]
    (1/2) >>> tensor([1, 2, 3, 4], dtype=torch.int32)
    (2/2) >>> tensor([5], dtype=torch.int32)
    >>> a = ht.zeros((4,5), split=0)
    (1/2) >>> tensor([[0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.]])
    (2/2) >>> tensor([[0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.]])
    >>> a[1:4, 1]
    (1/2) >>> tensor([0.])
    (2/2) >>> tensor([0., 0.])
    """
    # key can be: int, tuple, list, slice, DNDarray, torch tensor, numpy array, or sequence thereof

    if key is None:
        return self.expand_dims(0)
    if (
        key is ... or isinstance(key, slice) and key == slice(None)
    ):  # latter doesnt work with torch for 0-dim tensors
        return self

    original_split = self.split

    # Single-element indexing
    scalar = np.isscalar(key) or getattr(key, "ndim", 1) == 0
    if scalar:
        # single-element indexing on axis 0
        if self.ndim == 0:
            raise IndexError(
                "Too many indices for DNDarray: DNDarray is 0-dimensional, but 1 were indexed"
            )
        output_shape = self.gshape[1:]
        if original_split is None or original_split == 0:
            output_split = None
        else:
            output_split = original_split - 1
        split_key_is_ordered = 1
        out_is_balanced = True
        backwards_transpose_axes = tuple(range(self.ndim))
        key, root = self.__process_scalar_key(key, indexed_axis=0, return_local_indices=True)
        if root is None:
            # early out for single-element indexing not affecting split axis
            indexed_arr = self.larray[key]
            indexed_arr = DNDarray(
                indexed_arr,
                gshape=output_shape,
                dtype=self.dtype,
                split=output_split,
                device=self.device,
                comm=self.comm,
                balanced=out_is_balanced,
            )
            return indexed_arr
    else:
        # process multi-element key
        (
            self,
            key,
            output_shape,
            output_split,
            split_key_is_ordered,
            key_is_mask_like,
            out_is_balanced,
            root,
            backwards_transpose_axes,
        ) = self.__process_key(key, return_local_indices=True)

    if not self.is_distributed():
        # key is torch-proof, index underlying torch tensor
        indexed_arr = self.larray[key]
        # transpose array back if needed
        self = self.transpose(backwards_transpose_axes)
        return DNDarray(
            indexed_arr,
            gshape=output_shape,
            dtype=self.dtype,
            split=output_split,
            device=self.device,
            comm=self.comm,
            balanced=out_is_balanced,
        )

    if split_key_is_ordered == 1:
        if root is not None:
            # single-element indexing along split axis
            # prepare for Bcast: allocate buffer on all processes
            if self.comm.rank == root:
                indexed_arr = self.larray[key]
            else:
                indexed_arr = torch.zeros(
                    output_shape, dtype=self.larray.dtype, device=self.larray.device
                )
            # broadcast result to all processes
            self.comm.Bcast(indexed_arr, root=root)
            indexed_arr = DNDarray(
                indexed_arr,
                gshape=output_shape,
                dtype=self.dtype,
                split=output_split,
                device=self.device,
                comm=self.comm,
                balanced=out_is_balanced,
            )
            # transpose array back if needed
            self = self.transpose(backwards_transpose_axes)
            return indexed_arr

        # root is None, i.e. indexing does not affect split axis, apply as is
        indexed_arr = self.larray[key]
        # transpose array back if needed
        self = self.transpose(backwards_transpose_axes)
        return DNDarray(
            indexed_arr,
            gshape=output_shape,
            dtype=self.dtype,
            split=output_split,
            device=self.device,
            balanced=out_is_balanced,
            comm=self.comm,
        )

    # key along split axis is not ordered, indices are GLOBAL
    # prepare for communication of indices and data
    counts, displs = self.counts_displs()
    rank, size = self.comm.rank, self.comm.size

    key_is_single_tensor = isinstance(key, torch.Tensor)
    if key_is_single_tensor:
        split_key = key
    else:
        split_key = key[self.split]
    # split_key might be multi-dimensional, flatten it for communication
    if split_key.ndim > 1:
        original_split_key_shape = split_key.shape
        communication_split = output_split - (split_key.ndim - 1)
        split_key = split_key.flatten()
    else:
        communication_split = output_split

    # determine the number of elements to be received from each process
    recv_counts = torch.zeros((size, 1), dtype=torch.int64, device=self.larray.device)
    if key_is_mask_like:
        recv_indices = torch.zeros(
            (len(split_key), len(key)), dtype=split_key.dtype, device=self.larray.device
        )
    else:
        recv_indices = torch.zeros(
            (split_key.shape), dtype=split_key.dtype, device=self.larray.device
        )
    for p in range(size):
        cond1 = split_key >= displs[p]
        cond2 = split_key < displs[p] + counts[p]
        indices_from_p = torch.nonzero(cond1 & cond2, as_tuple=False)
        incoming_indices = split_key[indices_from_p].flatten()
        recv_counts[p, 0] = incoming_indices.numel()
        # store incoming indices in appropiate slice of recv_indices
        start = recv_counts[:p].sum().item()
        stop = start + recv_counts[p].item()
        if incoming_indices.numel() > 0:
            if key_is_mask_like:
                # apply selection to all dimensions
                for i in range(len(key)):
                    recv_indices[start:stop, i] = key[i][indices_from_p].flatten()
                recv_indices[start:stop, self.split] -= displs[p]
            else:
                recv_indices[start:stop] = incoming_indices - displs[p]
    # build communication matrix by sharing recv_counts with all processes
    # comm_matrix rows contain the send_counts for each process, columns contain the recv_counts
    comm_matrix = torch.zeros((size, size), dtype=torch.int64, device=self.larray.device)
    self.comm.Allgather(recv_counts, comm_matrix)
    send_counts = comm_matrix[:, rank]

    # active rank pairs:
    active_rank_pairs = torch.nonzero(comm_matrix, as_tuple=False)

    # Communication build-up:
    active_recv_indices_from = active_rank_pairs[torch.where(active_rank_pairs[:, 1] == rank)][
        :, 0
    ]
    active_send_indices_to = active_rank_pairs[torch.where(active_rank_pairs[:, 0] == rank)][
        :, 1
    ]
    rank_is_active = active_recv_indices_from.numel() > 0 or active_send_indices_to.numel() > 0

    # allocate recv_buf for incoming data
    recv_buf_shape = list(output_shape)
    if communication_split != output_split:
        # split key was flattened, flatten corresponding dims in recv_buf accordingly
        recv_buf_shape = (
            recv_buf_shape[:communication_split]
            + [recv_counts.sum().item()]
            + recv_buf_shape[output_split + 1 :]
        )
    else:
        recv_buf_shape[communication_split] = recv_counts.sum().item()
    recv_buf = torch.zeros(
        tuple(recv_buf_shape), dtype=self.larray.dtype, device=self.larray.device
    )
    if rank_is_active:
        # non-blocking send indices to `active_send_indices_to`
        send_requests = []
        for i in active_send_indices_to:
            start = recv_counts[:i].sum().item()
            stop = start + recv_counts[i].item()
            outgoing_indices = recv_indices[start:stop]
            send_requests.append(self.comm.Isend(outgoing_indices, dest=i))
            del outgoing_indices
        del recv_indices
        for i in active_recv_indices_from:
            # receive indices from `active_recv_indices_from`
            if key_is_mask_like:
                incoming_indices = torch.zeros(
                    (send_counts[i].item(), len(key)),
                    dtype=torch.int64,
                    device=self.larray.device,
                )
            else:
                incoming_indices = torch.zeros(
                    send_counts[i].item(), dtype=torch.int64, device=self.larray.device
                )
            self.comm.Recv(incoming_indices, source=i)
            # prepare send_buf for outgoing data
            if key_is_single_tensor:
                send_buf = self.larray[incoming_indices]
            else:
                if key_is_mask_like:
                    send_key = tuple(
                        incoming_indices[:, i].reshape(-1)
                        for i in range(incoming_indices.shape[1])
                    )
                    send_buf = self.larray[send_key]
                else:
                    send_key = list(key)
                    send_key[self.split] = incoming_indices
                    send_buf = self.larray[tuple(send_key)]
            # non-blocking send requested data to i
            send_requests.append(self.comm.Isend(send_buf, dest=i))
            del send_buf
        # allocate temporary recv_buf to receive data from all active processes
        tmp_recv_buf_shape = recv_buf_shape.copy()
        tmp_recv_buf_shape[communication_split] = recv_counts.max().item()
        tmp_recv_buf = torch.zeros(
            tuple(tmp_recv_buf_shape), dtype=self.larray.dtype, device=self.larray.device
        )
        for i in active_send_indices_to:
            # receive data from i
            tmp_recv_slice = [slice(None)] * tmp_recv_buf.ndim
            tmp_recv_slice[communication_split] = slice(0, recv_counts[i].item())
            self.comm.Recv(tmp_recv_buf[tmp_recv_slice], source=i)
            # write received data to appropriate portion of recv_buf
            cond1 = split_key >= displs[i]
            cond2 = split_key < displs[i] + counts[i]
            recv_buf_indices = torch.nonzero(cond1 & cond2, as_tuple=False).flatten()
            recv_buf_key = [slice(None)] * recv_buf.ndim
            recv_buf_key[communication_split] = recv_buf_indices
            recv_buf[recv_buf_key] = tmp_recv_buf[tmp_recv_slice]
        del tmp_recv_buf
        # wait for all non-blocking communication to finish
        for req in send_requests:
            req.Wait()
    if communication_split != output_split:
        # split_key has been flattened, bring back recv_buf to intended shape
        original_local_shape = (
            output_shape[:communication_split]
            + original_split_key_shape
            + output_shape[output_split + 1 :]
        )
        recv_buf = recv_buf.reshape(original_local_shape)

    # construct indexed array from recv_buf
    indexed_arr = DNDarray(
        recv_buf,
        gshape=output_shape,
        dtype=self.dtype,
        split=output_split,
        device=self.device,
        comm=self.comm,
        balanced=out_is_balanced,
    )
    # transpose array back if needed
    self = self.transpose(backwards_transpose_axes)
    return indexed_arr


from __future__ import annotations

from ._array_object import Array

import heat as ht


def sort(x: Array, /, *, axis: int = -1, descending: bool = False, stable: bool = True) -> Array:
    """
    Returns a sorted copy of an input array x.

    Parameters
    ----------
    x : Array
        Input array.
    axis : int
        Axis along which to sort. If set to ``-1``, the function must sort along the
        last axis. Default: ``-1``.
    descending : bool
        Sort order. If ``True``, the array must be sorted in descending order (by
        value). If ``False``, the array must be sorted in ascending order (by value).
        Default: ``False``.
    stable : bool
        Sort stability. If ``True``, the returned array maintains the relative order
        of ``x`` values which compare as equal. If ``False``, the returned array may
        or may not maintain the relative order of ``x`` values which compare as equal.
        Default: ``True``.
    """
    if stable:
        raise ValueError("Stable sorting not yet implemented")
    res = ht.sort(x._array, axis=axis, descending=descending)
    return Array._new(res[0])



"""
Linear Algebra Extension for the Array API standard.
"""

from __future__ import annotations

from ._dtypes import _numeric_dtypes, _result_type
from ._array_object import Array

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ._typing import Sequence, Tuple, Union

import heat as ht


def matmul(x1: Array, x2: Array, /) -> Array:
    """
    Computes the matrix product.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a numeric data type and at least one dimension.
    x2 : Array
        Second input array. Must have a numeric data type and at least one dimension.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in matmul")
    return Array._new(ht.matmul(x1._array, x2._array))


def matrix_transpose(x: Array, /) -> Array:
    """
    Transposes a matrix (or a stack of matrices) ``x``.

    Parameters
    ----------
    x : Array
        Input array having shape ``(..., M, N)`` and whose innermost two
        dimensions form ``MxN`` matrices.
    """
    if x.ndim < 2:
        raise ValueError("x must be at least 2-dimensional for matrix_transpose")
    return Array._new(ht.swapaxes(x._array, -1, -2))


def tensordot(
    x1: Array, x2: Array, /, *, axes: Union[int, Tuple[Sequence[int], Sequence[int]]] = 2
) -> Array:
    """
    Return a tensor contraction of ``x1`` and ``x2`` over specific axes.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a numeric data type.
    x2 : Array
        Second input array. Must have a numeric data type. Corresponding contracted axes of ``x1``
        and ``x2`` must be equal.
    axes : Union[int, Tuple[Sequence[int], Sequence[int]]]
        Number of axes (dimensions) to contract or explicit sequences of axes (dimensions) for
        ``x1`` and ``x2``, respectively. If ``axes`` is an ``int`` equal to ``N``, then contraction is
        performed over the last ``N`` axes of ``x1`` and the first ``N`` axes of ``x2`` in order.
        The size of each corresponding axis (dimension) must match. Must be nonnegative.
        If ``axes`` is a tuple of two sequences ``(x1_axes, x2_axes)``, the first sequence must apply
        to ``x1`` and the second sequence to ``x2``. Both sequences must have the same length.
        Each axis (dimension) ``x1_axes[i]`` for ``x1`` must have the same size as the respective axis
        (dimension) ``x2_axes[i]`` for ``x2``. Each sequence must consist of unique (nonnegative)
        integers that specify valid axes for each respective array.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in tensordot")
    return Array._new(ht.tensordot(x1._array, x2._array, axes=axes))


def vecdot(x1: Array, x2: Array, /, *, axis: int = -1) -> Array:
    """
    Computes the (vector) dot product of two arrays.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a numeric data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a numeric data type.
    axis : int
        Axis over which to compute the dot product. Must be an integer on the interval
        ``[-N, N)``, where ``N`` is the rank (number of dimensions) of the shape
        determined according to Broadcasting. If specified as a negative integer, the
        function determines the axis along which to compute the dot product by counting
        backward from the last dimension (where ``-1`` refers to the last dimension).
        By default, the function computes the dot product over the last axis.
        Default: ``-1``.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in vecdot")
    res = ht.vecdot(x1._array, x2._array, axis=axis)
    return Array._new(res.astype(_result_type(x1.dtype, x2.dtype)))



"""Module for TestCases used for heat"""

import os
import platform
import unittest
import tempfile

import numpy as np
import torch
from typing import Optional, Callable, Any, Union

import heat as ht
from heat.core import MPI, MPICommunication, dndarray, factories, types, Device
from heat.core.random import seed


# TODO adapt for GPU once this is working properly
class TestCase(unittest.TestCase):
    """Helper functions for unit tests"""

    __comm = MPICommunication()
    device: Device = ht.cpu
    _hostnames: Optional[list[str]] = None
    other_device: Optional[Device] = None
    envar: Optional[str] = None

    def setUp(self) -> None:
        """Sets initial RNG seed for testing"""
        seed(42)

    @classmethod
    def setUpClass(cls) -> None:
        """
        Read the environment variable 'HEAT_TEST_USE_DEVICE' and return the requested devices.
        Supported values
            - cpu: Use CPU only (default)
            - gpu: Use GPU only

        Raises
        ------
        RuntimeError if value of 'HEAT_TEST_USE_DEVICE' is not recognized

        """
        envar = os.getenv("HEAT_TEST_USE_DEVICE", "cpu")
        is_mps = False

        if envar == "cpu":
            ht.use_device("cpu")
            ht_device = ht.cpu
            other_device = ht.cpu
            if torch.cuda.is_available():
                torch.cuda.set_device(torch.device(ht.gpu.torch_device))
                other_device = ht.gpu
            elif torch.backends.mps.is_built() and torch.backends.mps.is_available():
                other_device = ht.gpu
        elif envar == "gpu":
            if torch.cuda.is_available():
                ht.use_device("gpu")
                torch.cuda.set_device(torch.device(ht.gpu.torch_device))
                ht_device = ht.gpu
                other_device = ht.cpu
            elif torch.backends.mps.is_built() and torch.backends.mps.is_available():
                ht.use_device("gpu")
                ht_device = ht.gpu
                other_device = ht.cpu
                is_mps = True
        else:
            raise RuntimeError(
                f"Value '{envar}' of environment variable 'HEAT_TEST_USE_DEVICE' is unsupported"
            )

        cls.device, cls.other_device, cls.envar, cls.is_mps = ht_device, other_device, envar, is_mps

    @property
    def comm(self) -> MPICommunication:
        """Returns the MPI communicator"""
        return self.__comm

    def get_rank(self) -> Optional[int]:
        """Returns the MPI rank"""
        return self.comm.rank

    def get_size(self) -> Optional[int]:
        """Returns the MPI size"""
        return self.comm.size

    @classmethod
    def get_hostnames(cls) -> list[str]:
        """Returns the name of the host machine(s)."""
        if not cls._hostnames:
            if platform.system() == "Windows":
                host = platform.uname().node
            else:
                host = os.uname()[1]
            cls._hostnames = list(set(cls.__comm.handle.allgather(host)))
        return cls._hostnames

    def assert_array_equal(
        self,
        heat_array: ht.DNDarray,
        expected_array: Union[np.ndarray, torch.Tensor],
        rtol: float = 1e-5,
        atol: float = 1e-08,
    ) -> None:
        """
        Check if the heat_array is equivalent to the expected_array. Therefore first the split heat_array is compared to
        the corresponding expected_array slice locally and second the heat_array is combined and fully compared with the
        expected_array.
        Note if the heat array is split it also needs to be balanced.

        Parameters
        ----------
        heat_array: heat.DNDarray
            The heat array which should be checked.
        expected_array: numpy.ndarray or torch.Tensor
            The array against which the heat_array should be checked.
        rtol: float
            The relative tolerance parameter.
        atol: float
            The absolute tolerance parameter.

        Raises
        ------
        AssertionError if the arrays do not equal.

        Examples
        --------
        >>> import numpy as np
        >>> import heat as ht
        >>> a = ht.ones((5, 5), split=1, dtype=ht.int32)
        >>> b = np.ones((5, 5), dtype=np.int32)
        >>> self.assert_array_equal(a, b)

        >>> c = np.ones((5, 5), dtype=np.int64)
        >>> self.assert_array_equal(a, c)
        AssertionError: [...]
        >>> c = np.zeros((5, 5), dtype=np.int32)
        >>> self.assert_array_equal(a, c)
        AssertionError: [...]
        """
        self._comm = heat_array.comm
        if isinstance(expected_array, torch.Tensor):
            # Does not work because heat sets an index while torch does not
            # self.assertEqual(expected_array.device, torch.device(heat_array.device.torch_device))
            expected_array = expected_array.cpu().numpy()

        self.assertIsInstance(
            heat_array,
            dndarray.DNDarray,
            f"The array to test was not a instance of ht.DNDarray. Instead got {type(heat_array)}.",
        )
        self.assertIsInstance(
            expected_array,
            np.ndarray,
            f"The array to test against was not a instance of numpy.ndarray or torch.Tensor Instead got {type(expected_array)}.",
        )
        self.assertEqual(
            heat_array.shape,
            expected_array.shape,
            f"Global shapes do not match. Got {heat_array.shape} expected {expected_array.shape}",
        )

        if not heat_array.is_balanced():
            # Array is not distributed correctly
            heat_array.balance_()

        split = heat_array.split
        offset, local_shape, slices = heat_array.comm.chunk(heat_array.gshape, split)
        self.assertEqual(
            heat_array.lshape,
            expected_array[slices].shape,
            f"Local shapes do not match. Got {heat_array.lshape} expected {expected_array[slices].shape}",
        )
        # compare local tensors to corresponding slice of expected_array
        is_allclose = torch.tensor(
            np.allclose(heat_array.larray.cpu(), expected_array[slices], atol=atol, rtol=rtol),
            dtype=torch.int32,
        )
        heat_array.comm.Allreduce(MPI.IN_PLACE, is_allclose, MPI.SUM)
        self.assertTrue(is_allclose == heat_array.comm.size)

    def assert_func_equal(
        self,
        shape: Union[tuple[Any, ...], list[Any]],
        heat_func: Callable[..., Any],
        numpy_func: Callable[..., Any],
        distributed_result: bool = True,
        heat_args: Optional[dict[str, Any]] = None,
        numpy_args: Optional[dict[str, Any]] = None,
        data_types: tuple[type, ...] = (np.int32, np.int64, np.float32, np.float64),
        low: int = -10000,
        high: int = 10000,
    ) -> None:
        """
        Creates random tensors of the given shape with different data types.
        All of these tensors will be tested with `ht.assert_func_equal_for_tensor`.

        Parameters
        ----------
        shape: tuple or list
            The shape of which a random tensors will be created and tested against
        heat_func: function
            The function that is to be tested
        numpy_func: function
            The numpy implementation of an equivalent function to test against
        heat_args: dictionary, optional
            The keyword arguments that will be passed to the heat function. Array and split function don't need to be
            specified. Default is {}.
        numpy_args: dictionary, optional
            The keyword arguments that will be passed to the numpy function. Array doesn't need to be specified.
            Default is {}.
        distributed_result: bool, optional
            Specify whether the result of the heat function is distributed across all nodes or all nodes have the full
            result. Default is True.
        data_types: list of numpy dtypes, optional
            Tensors with all of these dtypes will be created and tested. Each type must to be a numpy dtype.
            Default is [numpy.int32, numpy.int64, numpy.float32, numpy.float64]
        low: int, optional
            In case one of the data_types has integer types, this is the lower bound for the random values.
            Default is -10000
        high: int, optional
            In case one of the data_types has integer types, this is the upper bound for the random values.
            Default is 10000

        Raises
        ------
        AssertionError if the functions do not perform equally.

        Examples
        --------
        >>> import numpy as np
        >>> import heat as ht
        >>> self.assert_func_equal((2, 2), ht.exp, np.exp)

        >>> self.assert_func_equal((2, 2), ht.exp, np.log)
        AssertionError: [...]
        >>> self.assert_func_equal((1, 3, 5), ht.any, np.any, distributed_result=False)

        >>> heat_args = {"sorted": True, "axis": 0}
        >>> numpy_args = {"axis": 0}
        >>> self.assert_func_equal(
        ...     [5, 5, 5, 5], ht.unique, np.unique, heat_arg=heat_args, numpy_args=numpy_args
        ... )
        """
        if not isinstance(shape, tuple) and not isinstance(shape, list):
            raise ValueError(f"The shape must be either a list or a tuple but was {type(shape)}")

        if self.is_mps and np.float64 in data_types:
            # MPS does not support float64
            data_types = [dtype for dtype in data_types if dtype != np.float64]

        for dtype in data_types:
            tensor = self.__create_random_np_array(shape, dtype=dtype, low=low, high=high)
            self.assert_func_equal_for_tensor(
                tensor=tensor,
                heat_func=heat_func,
                numpy_func=numpy_func,
                heat_args=heat_args,
                numpy_args=numpy_args,
                distributed_result=distributed_result,
            )

    def assert_func_equal_for_tensor(
        self,
        tensor: Union[np.ndarray, torch.Tensor],
        heat_func: Callable[..., Any],
        numpy_func: Callable[..., Any],
        heat_args: Optional[dict[str, Any]] = None,
        numpy_args: Optional[dict[str, Any]] = None,
        distributed_result: bool = True,
    ) -> None:
        """
        Tests if the heat function and the numpy function create the equal result on the given tensor.

        Parameters
        ----------
        tensor: torch.Tensor or numpy.ndarray
            The tensor on which the heat function will be executed.
        heat_func: function
            The function that is to be tested
        numpy_func: function
            The numpy implementation of an equivalent function to test against
        heat_args: dictionary, optional
            The keyword arguments that will be passed to the heat function. Array and split function don't need to be
            specified. Default is {}.
        numpy_args: dictionary, optional
            The keyword arguments that will be passed to the numpy function. Array doesn't need to be specified.
            Default is {}.
        distributed_result: bool, optional
            Specify whether the result of the heat function is distributed across all nodes or all nodes have the full
            result. Default is True.

        Raises
        ------
        AssertionError if the functions to not perform equally.

        Examples
        --------
        >>> import numpy as np
        >>> import heat as ht
        >>> a = np.arange(10)
        >>> self.assert_func_equal_for_tensor(a, ht.exp, np.exp)

        >>> self.assert_func_equal_for_tensor(a, ht.exp, np.log)
        AssertionError: [...]
        >>> self.assert_func_equal_for_tensor(a, ht.any, np.any, distributed_result=False)

        >>> a = torch.ones([5, 5, 5, 5])
        >>> heat_args = {"sorted": True, "axis": 0}
        >>> numpy_args = {"axis": 0}
        >>> self.assert_func_equal_for_tensor(
        ...     a, ht.unique, np.unique, heat_arg=heat_args, numpy_args=numpy_args
        ... )
        """
        self.assertTrue(callable(heat_func))
        self.assertTrue(callable(numpy_func))

        if heat_args is None:
            heat_args = {}
        if numpy_args is None:
            numpy_args = {}

        if isinstance(tensor, np.ndarray):
            torch_tensor = torch.from_numpy(tensor.copy())
            torch_tensor = torch_tensor.to(self.device.torch_device)
            np_array = tensor
        elif isinstance(tensor, torch.Tensor):
            torch_tensor = tensor
            np_array = tensor.cpu().numpy().copy()
        else:
            raise TypeError(
                f"The input tensors type must be one of [tuple, list, numpy.ndarray, torch.tensor] but is {type(tensor)}"
            )

        dtype = types.canonical_heat_type(torch_tensor.dtype)
        np_res = numpy_func(np_array, **numpy_args)
        if not isinstance(np_res, np.ndarray):
            np_res = np.array(np_res)

        for i in range(len(tensor.shape)):
            ht_array = factories.array(
                torch_tensor, split=i, dtype=dtype, device=self.device, comm=self.comm
            )
            ht_res = heat_func(ht_array, **heat_args)

            self.assertEqual(ht_array.device, ht_res.device)
            self.assertEqual(ht_array.larray.device, ht_res.larray.device)
            if distributed_result:
                self.assert_array_equal(ht_res, np_res)
            else:
                self.assertTrue(np.array_equal(ht_res.larray.cpu().numpy(), np_res))

    def assertTrue_memory_layout(self, tensor: ht.DNDarray, order: str) -> None:
        """
        Checks that the memory layout of a given heat tensor is as specified by argument order.

        Parameters
        ----------
        tensor: DNDarray
            The input array.
        order: str
            'C' for C-like (row-major), 'F' for Fortran-like (column-major) memory layout.
        """
        stride = tensor.larray.stride()
        row_major = all(np.diff(list(stride)) <= 0)
        column_major = all(np.diff(list(stride)) >= 0)
        if order == "C":
            return self.assertTrue(row_major)
        elif order == "F":
            return self.assertTrue(column_major)
        else:
            raise ValueError(f"expected order to be 'C' or 'F', but was {order}")

    def __create_random_np_array(
        self,
        shape: Union[list[Any], tuple[Any]],
        dtype: type = np.float32,
        low: int = -10000,
        high: int = 10000,
    ) -> np.ndarray:
        """
        Creates a random array based on the input parameters.
        The used seed will be printed to stdout for debugging purposes.

        Parameters
        ----------
        shape: list or tuple
            The shape of the random array to be created.
        dtype: np.dtype, optional
            The datatype of the resulting array.
            If dtype is subclass of np.floating then numpy.random.randn is used to create random values.
            If dtype is subclass of np.integer then numpy.random.randint is called with the low and high argument to
            create the random values.
            Default is numpy.float64
        low: int, optional
            In case dtype is an integer type, this is the lower bound for the random values.
            Default is -10000
        high: int, optional
            In case dtype is an integer type, this is the upper bound for the random values.
            Default is 10000

        Returns
        -------
        res: numpy.ndarray
            An array of random values with the specified shape and dtype.

        Raises
        ------
        ValueError if the dtype is not a subtype of numpy.integer or numpy.floating
        """
        seed = np.random.randint(1000000, size=(1,))
        # print("using seed {} for random values".format(seed))
        self.comm.Bcast(seed, root=0)
        np.random.seed(seed=seed.item())
        if issubclass(dtype, np.floating):
            array: np.ndarray = np.random.randn(*shape)
        elif issubclass(dtype, np.integer):
            array = np.random.randint(low=low, high=high, size=shape)
        else:
            raise ValueError(
                f"Unsupported dtype. Expected a subclass of `np.floating` or `np.integer` but got {dtype}"
            )
        array = array.astype(dtype)
        return array

    def get_tmpdir(self):
        """
        Create a temporary directory in the current working directory.
        When running with multiple tasks, only rank 0 creates and destroys the directory, but the path is communicated across all ranks.
        The directory is cleaned up / deleted when the variable tmpdir is garbage collected on rank 0.

        Returns
        -------
        path : str
            Path to the temporary directory
        tmpdir : object or None
            On rank 0: A Python object administering the temporary directory, on other ranks, None.
        """
        if ht.MPI_WORLD.rank == 0:
            tmpdir = tempfile.TemporaryDirectory(dir="./")
            path = tmpdir.name
        else:
            tmpdir = None
            path = None

        path = ht.MPI_WORLD.bcast(path, root=0)
        return path, tmpdir  # need to return tmpdir here so that its not cleaned up at this point



import heat.core.devices as ht_devices

from types import SimpleNamespace
from heat.core.types import (
    bool,
    complex64,
    complex128,
    float16,
    float32,
    float64,
    int8,
    int16,
    int32,
    int64,
    uint8,
)


def __array_namespace_info__():
    """Returns a namespace with Array API namespace inspection utilities."""
    info = SimpleNamespace()
    info.capabilities = capabilities
    info.default_device = default_device
    info.default_dtypes = default_dtypes
    info.devices = devices
    info.dtypes = dtypes
    return info


def capabilities():
    """Returns a dictionary of array library capabilities."""
    return {"boolean indexing": True, "data-dependent shapes": True, "max dimensions": 64}


def default_device():
    """Returns the default device."""
    return ht_devices.get_device()


def default_dtypes(*, device=None):
    """Returns a dictionary containing default data types."""
    if device is None:
        device = default_device()

    if not isinstance(device, ht_devices.Device):
        raise ValueError(f"Device not understood: {device}")

    if "mps" in device.torch_device:
        return {
            "real floating": float32,
            "complex floating": complex64,
            "integral": int32,
            "indexing": int32,
        }

    return {
        "real floating": float32,
        "complex floating": complex64,
        "integral": int64,
        "indexing": int64,
    }

    raise ValueError(f"Unsupported device: {device}")


def devices():
    """Returns a list of supported devices which are available at runtime."""
    if hasattr(ht_devices, "gpu"):
        return (ht_devices.cpu, ht_devices.gpu)
    else:
        return (ht_devices.cpu,)


def dtypes(*, device=None, kind=None):
    """Returns a dictionary of supported Array API data types"""
    if device is None:
        device = default_device()

    if not isinstance(device, ht_devices.Device):
        raise ValueError(f"Device not understood: {device}")

    if "mps" in device.torch_device:
        if kind is None:
            return {
                "bool": bool,
                "int8": int8,
                "int16": int16,
                "int32": int32,
                "uint8": uint8,
                "float32": float32,
                "complex64": complex64,
            }
        if kind == "bool":
            return {
                "bool": bool,
            }
        if kind == "signed integer":
            return {
                "int8": int8,
                "int16": int16,
                "int32": int32,
            }
        if kind == "unsigned integer":
            return {
                "uint8": uint8,
            }
        if kind == "integral":
            return {
                "int8": int8,
                "int16": int16,
                "int32": int32,
                "uint8": uint8,
            }
        if kind == "real floating":
            return {
                "float32": float32,
            }
        if kind == "complex floating":
            return {
                "complex64": complex64,
            }
        if kind == "numeric":
            return {
                "int8": int8,
                "int16": int16,
                "int32": int32,
                "uint8": uint8,
                "float32": float32,
                "complex64": complex64,
            }

    if kind is None:
        return {
            "bool": bool,
            "int8": int8,
            "int16": int16,
            "int32": int32,
            "int64": int64,
            "uint8": uint8,
            "float32": float32,
            "float64": float64,
            "complex64": complex64,
            "complex128": complex128,
        }
    if kind == "bool":
        return {
            "bool": bool,
        }
    if kind == "signed integer":
        return {
            "int8": int8,
            "int16": int16,
            "int32": int32,
            "int64": int64,
        }
    if kind == "unsigned integer":
        return {
            "uint8": uint8,
        }
    if kind == "integral":
        return {
            "int8": int8,
            "int16": int16,
            "int32": int32,
            "int64": int64,
            "uint8": uint8,
        }
    if kind == "real floating":
        return {
            "float32": float32,
            "float64": float64,
        }
    if kind == "complex floating":
        return {
            "complex64": complex64,
            "complex128": complex128,
        }
    if kind == "numeric":
        return {
            "int8": int8,
            "int16": int16,
            "int32": int32,
            "int64": int64,
            "uint8": uint8,
            "float32": float32,
            "float64": float64,
            "complex64": complex64,
            "complex128": complex128,
        }
    if isinstance(kind, tuple):
        res = {}
        for k in kind:
            res |= dtypes(device=device, kind=k)
        return res
    raise ValueError(f"Unsupported kind: {kind}")



"""
Heat command line interface module.
"""

import torch
import platform
import mpi4py
import argparse

from heat.core.version import __version__ as ht_version
from heat.core._config import CUDA_AWARE_MPI, ROCM_AWARE_MPI, GPU_AWARE_MPI


def cli() -> None:
    """
    Command line interface entrypoint.
    """
    parser = argparse.ArgumentParser(
        prog="heat", description="Commmand line utilities of the Helmholtz Analytics Toolkit"
    )
    parser.add_argument(
        "-i", "--info", action="store_true", help="Print version and platform information"
    )

    args = parser.parse_args()
    if args.info:
        plaform_info()
    else:
        parser.print_help()


def plaform_info():
    """
    Print the current software stack being used by heat, including available devices.
    """
    print("HeAT: Helmholtz Analytics Toolkit")
    print(f"  Version: {ht_version}")
    print(f"  Platform: {platform.platform()}")

    print(f"  mpi4py Version: {mpi4py.__version__}")
    print(f"  MPI Library Version: {mpi4py.MPI.Get_library_version()}")

    print(f"  Torch Version: {torch.__version__}")
    print(f"  CUDA Available: {torch.cuda.is_available()}")
    if torch.cuda.is_available():
        def_device = torch.cuda.current_device()
        print(f"    Device count: {torch.cuda.device_count()}")
        print(f"    Default device: {def_device}")
        print(f"    Device name: {torch.cuda.get_device_name(def_device)}")
        print(f"    Device name: {torch.cuda.get_device_properties(def_device)}")
        print(
            f"    Device memory: {torch.cuda.get_device_properties(def_device).total_memory / 1024**3} GiB"
        )
        print(f"    CUDA Aware MPI: {CUDA_AWARE_MPI}")
        print(f"    ROCM Aware MPI: {ROCM_AWARE_MPI}")
        print(f"    GPU Aware MPI: {GPU_AWARE_MPI}")



"""
Vmap module.
This implements a functionality similar to PyTorchs vmap function.
Requires PyTorch 2.0.0 or higher.
"""

import torch

from .dndarray import DNDarray
from .factories import array
from .communication import MPI_WORLD
from typing import Union, Tuple, Optional, Callable

__all__ = ["vmap"]


def vmap(
    func: Callable[[Tuple[torch.Tensor]], Tuple[torch.Tensor]],
    out_dims: Union[Tuple[int], int] = 0,
    randomness: str = "error",
    *,
    chunk_size: int = None,
) -> Callable[[Tuple[DNDarray]], Tuple[DNDarray]]:
    """
    Apply a function to a DNDarray in a vectorized way.
    `heat.vmap` return a callable that can be applied to DNDarrays.
    Vectorization will automatically take place along the split axis/axes of the DNDarray(s);
    therefore, unlike in PyTorch, there is no argument `in_dims`.
    What we here refer to as "split axis/dimension" in the Heat terminology is often referred to as "batch axis/dimension" in the PyTorch terminology.

    Parameters
    ----------
    func : callable
        The function to apply in a vmapped way to the DNDarray(s). It must take PyTorch tensor(s) as positional arguments.
        Additional parameters, not to be vmapped over, can be passed as keyword arguments. The callable returned by
        by `heat.vmap` will also accept these keyword arguments.
    out_dims : int or tuple of int, optional
        The dimensions of the output(s) that are mapped over; identical to the split dimension(s) of the output(s).
        Default is 0.
    randomness : {'error', 'different', 'same'}, optional
        Determines how to handle randomness in the function to be vmapped. This argument is directly passed to the underlying PyTorch vmaps;
        see the corresponding PyTorch documentation for more information and the note below.
        If 'error' (default), an error is raised if the function to be mapped contains randomness.
    chunk_size : int, optional
        The size of the chunks to use for the process-local computation.
        If None (default), apply a single PyTorch vmap over the process-local chunks of data. If not None, then compute the process-local PyTorch vmap `chunk_size`
        many samples at a time. Note that `chunk_size=1` is equivalent to computing the process-local PyTorch vmap's with a for-loop.
        If you run into memory issues computing the vmap, please try a non-None chunk_size.

    Note
    ------
    This function is a wrapper around PyTorch's `torch.vmap` function. In essence, a PyTorch vmap is applied to the input function `func` on each MPI process separately.
    This process-local PyTorch-vmapped function is then applied to the process-local chunks of the input DNDarray(s).

    Please note that the options 'same' and 'different' for `randomness` will result in behaviour different from the one known by PyTorch as (at least currently)
    no actions are taken to synchronize randomness across the MPI processes.
    """
    # check PyTorch version, return error if not 2.0.0 or higher
    if torch.__version__ < "2.0.0":
        raise RuntimeError("The function `heat.vmap` requires PyTorch 2.0.0 or higher.")
    # rough check of input argument types
    if not callable(func):
        raise TypeError("The input function `func` must be callable.")
    if randomness not in ["error", "different", "same"]:
        raise ValueError(
            "The input argument `randomness` must be one of the strings 'error', 'different', or 'same'."
        )
    if chunk_size is not None and not isinstance(chunk_size, int):
        raise TypeError("The input argument `chunk_size` must be None or an integer.")
    else:
        if chunk_size is not None and chunk_size < 1:
            raise ValueError("If an integer, the input argument `chunk_size` must be at least 1.")

    def vmapped_func(*args, **kwargs):
        for arg in args:
            if not isinstance(arg, DNDarray):
                raise TypeError(
                    f"All inputs to the vmapped-version of your function must be DNDarrays, but one is {type(arg)}."
                )
        in_dims = tuple([arg.split for arg in args])

        # apply Torch vmap to the input function and the result to the local arrays of the input DNDarray
        torch_vmap_func = torch.vmap(
            func, in_dims, out_dims, randomness=randomness, chunk_size=chunk_size
        )
        out_larrays = torch_vmap_func(*[arg.larray for arg in args], **kwargs)

        if isinstance(out_larrays, torch.Tensor):
            # circumvent misinterpretation of the following call of len() in case of a single output
            out_larrays = [out_larrays]
        if isinstance(out_dims, int):
            out_split = [out_dims] * len(out_larrays)
        else:
            out_split = out_dims
            if len(out_split) != len(out_larrays):
                raise ValueError(
                    f"The number of output DNDarrays ({len(out_larrays)}) must match the number of their split dimensions provided in `out_dims` ({len(out_split)})."
                )
        # generate output DNDarray(s)
        out_dndarrays = [
            array(out_larrays[k], is_split=out_split[k]) for k in range(len(out_larrays))
        ]
        return tuple(out_dndarrays)

    return vmapped_func



"""
Module implementing decomposition techniques, such as PCA.
"""

import heat as ht
from typing import Optional, Tuple, Union
from ..core.linalg.svdtools import _isvd

if ht.io.supports_hdf5():
    import h5py

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

"""
The implementation is heavily inspired by the corresponding routines in scikit-learn (https://scikit-learn.org/stable/modules/decomposition.html).
Please note that sometimes deviations cannot be avoided due to different algorithms and the distributed nature of the heat package.
"""


class PCA(ht.TransformMixin, ht.BaseEstimator):
    """
    Pricipal Component Analysis (PCA).

    Linear dimensionality reduction using Singular Value Decomposition of the data to project it to a lower dimensional space.
    The input data is centered but not scaled for each feature before applying the SVD.

    Parameters
    ----------
    n_components : int, float, None, default=None
        Number of components to keep. If n_components is not set all components are kept.
        If n_components is an integer, it specifies the number of components to keep.
        If n_components is a float between 0 and 1, it specifies the fraction of variance explained by the components to keep.
    copy : bool, default=True
        In-place operations are not yet supported. Please set copy=True.
    whiten : bool, default=False
        Not yet supported.
    svd_solver : {'full', 'hierarchical'}, default='hierarchical'
        'full' : Full SVD is performed. In general, this is more accurate, but also slower. So far, this is only supported for tall-skinny or short-fat data.
        'hierarchical' : Hierarchical SVD, i.e., an algorithm for computing an approximate, truncated SVD, is performed. Only available for data split along axis no. 0.
        'randomized' : Randomized SVD is performed.
    tol : float, default=None
        Not yet necessary as iterative methods for PCA are not yet implemented.
    iterated_power :  int, default=0
        if svd_solver='randomized', this parameter is the number of iterations for the power method.
        Choosing `iterated_power > 0` can lead to better results in the case of slowly decaying singular values but is computationally more expensive.
    n_oversamples : int, default=10
        if svd_solver='randomized', this parameter is the number of additional random vectors to sample the range of X so that the range of X can be approximated more accurately.
    power_iteration_normalizer : {'qr'}, default='qr'
        if svd_solver='randomized', this parameter is the normalization form of the iterated power method. So far, only QR is supported.
    random_state : int, default=None
        if svd_solver='randomized', this parameter allows to set the seed for the random number generator.

    Attributes
    ----------
    components_ : DNDarray of shape (n_components, n_features)
        Principal axes in feature space, representing the directions of maximum variance in the data. The components are sorted by explained_variance_.
    explained_variance_ : DNDarray of shape (n_components,)
        The amount of variance explained by each of the selected components.
        Not supported by svd_solver='hierarchical' and svd_solver='randomized'.
    explained_variance_ratio_ : DNDarray of shape (n_components,)
        Percentage of variance explained by each of the selected components.
        Not supported by svd_solver='hierarchical' and svd_solver='randomized'.
    total_explained_variance_ratio_ : float
        The percentage of total variance explained by the selected components together.
        For svd_solver='hierarchical', an lower estimate for this quantity is provided; see :func:`ht.linalg.hsvd_rtol` and :func:`ht.linalg.hsvd_rank` for details.
        Not supported by svd_solver='randomized'.
    singular_values_ : DNDarray of shape (n_components,)
        The singular values corresponding to each of the selected components.
        Not supported by svd_solver='hierarchical' and svd_solver='randomized'.
    mean_ : DNDarray of shape (n_features,)
        Per-feature empirical mean, estimated from the training set.
    n_components_ : int
        The estimated number of components.
    n_samples_ : int
        Number of samples in the training data.
    noise_variance_ : float
        not yet implemented

    Notes
    -----
    Hierarchical SVD (`svd_solver = "hierarchical"`) computes an approximate, truncated SVD. Thus, the results are not exact, in general, unless `n_components`
    chosen is larger than the actual rank (=matrix rank) of the underlying data; see :func:`ht.linalg.hsvd_rank` and :func:`ht.linalg.hsvd_rtol` for details.
    Randomized SVD (`svd_solver = "randomized"`) is a stochastic algorithm that computes an approximate, truncated SVD.
    """

    def __init__(
        self,
        n_components: Optional[Union[int, float]] = None,
        copy: bool = True,
        whiten: bool = False,
        svd_solver: str = "hierarchical",
        tol: Optional[float] = None,
        iterated_power: Union[str, int] = 0,
        n_oversamples: int = 10,
        power_iteration_normalizer: str = "qr",
        random_state: Optional[int] = None,
    ):
        # check correctness of inputs
        if not copy:
            raise NotImplementedError(
                "In-place operations for PCA are not supported at the moment. Please set copy=True."
            )
        if whiten:
            raise NotImplementedError("Whitening is not yet supported. Please set whiten=False.")
        if not (svd_solver == "full" or svd_solver == "hierarchical" or svd_solver == "randomized"):
            raise ValueError(
                "At the moment, only svd_solver='full' (for tall-skinny or short-fat data), svd_solver='hierarchical', and svd_solver='randomized' are supported. \n An implementation of the 'full' option for arbitrarily shaped data is already planned."
            )
        if not isinstance(iterated_power, int):
            raise TypeError(
                "iterated_power must be an integer. The option 'auto' is not yet supported."
            )
        if isinstance(iterated_power, int) and iterated_power < 0:
            raise ValueError("if an integer, iterated_power must be greater or equal to 0.")
        if power_iteration_normalizer != "qr":
            raise ValueError("Only power_iteration_normalizer='qr' is supported yet.")
        if not isinstance(n_oversamples, int) or n_oversamples < 0:
            raise ValueError("n_oversamples must be a non-negative integer.")
        if tol is not None:
            raise ValueError(
                "Argument tol is not yet necessary as iterative methods for PCA are not yet implemented. Please set tol=None."
            )
        if random_state is not None and not isinstance(random_state, int):
            raise ValueError(f"random_state must be None or an integer, was {type(random_state)}.")
        if (
            n_components is not None
            and not (isinstance(n_components, int) and n_components >= 1)
            and not (isinstance(n_components, float) and n_components > 0.0 and n_components < 1.0)
        ):
            raise ValueError(
                "n_components must be None, in integer greater or equal to 1 or a float in (0,1). Option 'mle' is not supported at the moment."
            )

        self.n_components = n_components
        self.copy = copy
        self.whiten = whiten
        self.svd_solver = svd_solver
        self.tol = tol
        self.iterated_power = iterated_power
        self.n_oversamples = n_oversamples
        self.power_iteration_normalizer = power_iteration_normalizer
        self.random_state = random_state
        if self.random_state is not None:
            # set random seed accordingly
            ht.random.seed(self.random_state)

        # set future attributes to None to initialize those that will not be computed later on with None (e.g., explained_variance_ for svd_solver='hierarchical')
        self.components_ = None
        self.explained_variance_ = None
        self.explained_variance_ratio_ = None
        self.total_explained_variance_ratio_ = None
        self.singular_values_ = None
        self.mean_ = None
        self.n_components_ = None
        self.n_samples_ = None
        self.noise_variance_ = None

    def fit(self, X: ht.DNDarray, y=None) -> Self:
        """
        Fit the PCA model with data X.

        Parameters
        ----------
        X : DNDarray of shape (n_samples, n_features)
            Data set of which PCA has to be computed.
        y : Ignored
            Not used, present for API consistency by convention.
        """
        ht.sanitize_in(X)
        if y is not None:
            raise ValueError(
                "Argument y is ignored and just present for API consistency by convention."
            )

        # center data
        self.mean_ = X.mean(axis=0)
        X_centered = X - self.mean_

        # set n_components
        # note: n_components is an argument passed by the user to the PCA object and encodes the TARGET number of components
        #       n_components_ is an attribute of the PCA object that stores the ACTUAL number of components
        if self.n_components is None:
            self.n_components_ = min(X.shape)
        else:
            self.n_components_ = self.n_components

        # compute SVD via "full" SVD
        if self.svd_solver == "full" or not X.is_distributed():
            _, S, Vt = ht.linalg.svd(X_centered, full_matrices=False)
            total_variance = (S**2).sum() / (X.shape[0] - 1)
            if not isinstance(self.n_components_, int):
                # truncation w.r.t. prescribed bound on explained variance
                # determine n_components_ accordingly
                explained_variance_threshold = self.n_components_ * total_variance.larray.item()
                explained_variance_cumsum = (S**2).larray.cumsum(0) / (X.shape[0] - 1)
                self.n_components_ = (
                    len(
                        explained_variance_cumsum[
                            explained_variance_cumsum <= explained_variance_threshold
                        ]
                    )
                    + 1
                )
            self.components_ = Vt[: self.n_components_, :]
            self.singular_values_ = S[: self.n_components_]
            self.explained_variance_ = (S**2)[: self.n_components_] / (X.shape[0] - 1)
            self.explained_variance_ratio_ = self.explained_variance_ / total_variance
            self.total_explained_variance_ratio_ = self.explained_variance_ratio_.sum().item()
        # compute SVD via "hierarchical" SVD
        elif self.svd_solver == "hierarchical":
            if X.split != 0:
                raise ValueError(
                    "PCA with hierarchical SVD is only available for data split along axis 0."
                )
            if isinstance(self.n_components_, float):
                # hierarchical SVD with prescribed upper bound on relative error
                # note: "upper bound on relative error" (hsvd_rtol) is "1 - lower bound" (PCA)
                _, S, Vt, info = ht.linalg.hsvd_rtol(
                    X_centered, (1 - self.n_components_) ** 0.5, compute_sv=True, safetyshift=0
                )
            else:
                # hierarchical SVD with prescribed, fixed (truncation) rank
                _, S, Vt, info = ht.linalg.hsvd_rank(
                    X_centered, self.n_components_, compute_sv=True, safetyshift=0
                )
            self.n_components_ = Vt.shape[0]
            self.components_ = Vt
            self.total_explained_variance_ratio_ = 1 - info.larray.item() ** 2

        else:
            # compute SVD via "randomized" SVD
            _, S, Vt = ht.linalg.rsvd(
                X_centered,
                self.n_components_,
                n_oversamples=self.n_oversamples,
                power_iter=self.iterated_power,
            )
            self.components_ = Vt
            self.n_components_ = Vt.shape[0]

        self.n_samples_ = X.shape[0]
        self.noise_variance_ = None  # not yet implemented

        return self

    def transform(self, X: ht.DNDarray) -> ht.DNDarray:
        """
        Apply dimensionality based on PCA to X.

        Parameters
        ----------
        X : DNDarray of shape (n_samples, n_features)
            Data set to be transformed.
        """
        ht.sanitize_in(X)
        if X.shape[1] != self.mean_.shape[0]:
            raise ValueError(
                f"X must have the same number of features as the training data. Expected {self.mean_.shape[0]} but got {X.shape[1]}."
            )

        # center data and apply PCA
        X_centered = X - self.mean_
        return X_centered @ self.components_.T

    def inverse_transform(self, X: ht.DNDarray) -> ht.DNDarray:
        """
        Transform data back to its original space.

        Parameters
        ----------
        X : DNDarray of shape (n_samples, n_components)
            Data set to be transformed back.
        """
        ht.sanitize_in(X)
        if X.shape[1] != self.n_components_:
            raise ValueError(
                f"Dimension mismatch. Expected input of shape n_points x {self.n_components_} but got {X.shape}."
            )

        return X @ self.components_ + self.mean_


class IncrementalPCA(ht.TransformMixin, ht.BaseEstimator):
    """
    Incremental Principal Component Analysis (PCA).

    This class allows for incremental updates of the PCA model. This is especially useful for large data sets that do not fit into memory.

    An example how to apply this class is given in, e.g., `benchmarks/cb/decomposition.py`.

    Parameters
    ----------
    n_components : int, optional
        Number of components to keep. If `n_components` is not set all components are kept (default).
    copy : bool, default=True
        In-place operations are not yet supported. Please set `copy=True`.
    whiten : bool, default=False
        Not yet supported.
    batch_size : int, optional
        Currently not needed and only added for API consistency and possible future extensions.

    Attributes
    ----------
    components_ : DNDarray of shape (n_components, n_features)
        Principal axes in feature space, representing the directions of maximum variance in the data. The components are sorted by `explained_variance_.
    singular_values_ : DNDarray of shape (n_components,)
        The singular values corresponding to each of the selected components.
    mean_ : DNDarray of shape (n_features,)
        Per-feature empirical mean, estimated from the training set.
    n_components_ : int
        The estimated number of components.
    n_samples_seen_ : int
        Number of samples processed so far.
    """

    def __init__(
        self,
        n_components: Optional[int] = None,
        copy: bool = True,
        whiten: bool = False,
        batch_size: Optional[int] = None,
    ):
        if not copy:
            raise NotImplementedError(
                "In-place operations for PCA are not supported at the moment. Please set copy=True."
            )
        if whiten:
            raise NotImplementedError("Whitening is not yet supported. Please set whiten=False.")
        if n_components is not None:
            if not isinstance(n_components, int):
                raise TypeError(
                    f"n_components must be None or an integer, but is {type(n_components)}."
                )
            else:
                if n_components < 1:
                    raise ValueError("if an integer, n_components must be greater or equal to 1.")
        self.whiten = whiten
        self.n_components = n_components
        self.batch_size = batch_size
        self.components_ = None
        # self.explained_variance_ = None            # not yet supported
        # self.explained_variance_ratio_ = None      # not yet supported
        self.singular_values_ = None
        self.mean_ = None
        self.n_components_ = None
        self.batch_size_ = None
        self.n_samples_seen_ = 0

    def fit(self, path: str, chunk_size: int, dataset: str = "DATA") -> Self:
        """
        Fit the IncrementalPCA model using data loaded in chunks from a HDF5 file.

        This method processes data incrementally, loading chunks of data from a file and updating the PCA model iteratively.
        It is particularly useful for large datasets that cannot fit into memory.

        Parameters
        ----------
        path : str
            Path to the file containing the dataset. The file must be in HDF5 format.
        chunk_size : int
            Number of rows to load and process in each chunk. Must be smaller than or equal to the total number of rows in the dataset.
        dataset : str, default="DATA"
            Name of the dataset within the file to load.

        Returns
        -------
        Self
            The fitted IncrementalPCA instance.

        Raises
        ------
        ValueError
            If the file format is not HDF5.
            If `chunk_size` is larger than the number of rows in the dataset.
            If the number of columns is smaller than the number of processes.
        """
        if not ht.io.supports_hdf5():
            raise RuntimeError(
                "Computing IncrementalPCA from an HDF5 file requires HDF5 support, which is not available. Please install heat with HDF5 support."
            )
        if path.endswith(".h5"):
            with h5py.File(path, "r") as f:
                shape = f[dataset].shape

            # This additional intermediate step makes it easier to implement support for other file formats (such as Zarr) later on.
            def load_data_from_file(start, end):
                return ht.load_hdf5(
                    path, dataset=dataset, split=0, slices=[slice(start, end), None]
                )
        else:
            raise ValueError("`path` must direct to a HDF5 file.")

        # Check if the number of columns is at least equal to the number of processes
        if shape[1] < ht.MPI_WORLD.size:
            raise ValueError(
                f"The number of columns ({shape[1]}) must be at least equal to the number of processes ({ht.MPI_WORLD.size})."
            )

        if chunk_size > shape[0]:
            raise ValueError(
                "The `chunk_size` cannot be bigger than the number of rows of the chosen file."
            )

        start = 0
        end = chunk_size

        while start < shape[0]:
            if end <= shape[0]:
                X = load_data_from_file(start, end)
            else:
                X = load_data_from_file(start, shape[0])
            self.partial_fit(X)
            start += chunk_size
            end += chunk_size

    def partial_fit(self, X: ht.DNDarray, y=None):
        """
        One single step of incrementally building up the PCA.
        Input X is the current batch of data that needs to be added to the existing PCA.
        """
        ht.sanitize_in(X)
        if y is not None:
            raise ValueError(
                "Argument y is ignored and just present for API consistency by convention."
            )
        if self.n_samples_seen_ == 0:
            # this is the first batch of data, hence we need to initialize everything
            if self.n_components is None:
                self.n_components_ = min(X.shape)
            else:
                self.n_components_ = min(X.shape[0], X.shape[1], self.n_components)

            self.mean_ = X.mean(axis=0)
            X_centered = X - self.mean_
            _, S, Vt = ht.linalg.svd(X_centered)
            self.components_ = Vt[: self.n_components_, :]
            self.singular_values_ = S[: self.n_components_]
            self.n_samples_seen_ = X.shape[0]

        else:
            # if already batches of data have been seen before, only an update is necessary
            U, S, mean = _isvd(
                X.T,
                self.components_.T,
                self.singular_values_,
                Vt_old=None,
                maxrank=self.n_components,
                old_matrix_size=self.n_samples_seen_,
                old_rowwise_mean=self.mean_,
            )
            self.components_ = U.T
            self.singular_values_ = S
            self.mean_ = mean
            self.n_samples_seen_ += X.shape[0]
            self.n_components_ = self.components_.shape[0]

    def transform(self, X: ht.DNDarray) -> ht.DNDarray:
        """
        Apply dimensionality based on PCA to X.

        Parameters
        ----------
        X : DNDarray of shape (n_samples, n_features)
            Data set to be transformed.
        """
        ht.sanitize_in(X)
        if X.shape[1] != self.mean_.shape[0]:
            raise ValueError(
                f"X must have the same number of features as the training data. Expected {self.mean_.shape[0]} but got {X.shape[1]}."
            )

        # center data and apply PCA
        X_centered = X - self.mean_
        return X_centered @ self.components_.T

    def inverse_transform(self, X: ht.DNDarray) -> ht.DNDarray:
        """
        Transform data back to its original space.

        Parameters
        ----------
        X : DNDarray of shape (n_samples, n_components)
            Data set to be transformed back.
        """
        ht.sanitize_in(X)
        if X.shape[1] != self.n_components_:
            raise ValueError(
                f"Dimension mismatch. Expected input of shape n_points x {self.n_components_} but got {X.shape}."
            )

        return X @ self.components_ + self.mean_



"""
Randomized linear algebra algorithms, including randomized SVD (rSVD) and randomized EIGH (reigh).
"""

import torch
from typing import Union, Tuple

from ..dndarray import DNDarray
from ..linalg import matmul, qr, svd
from ..random import randn
from ..sanitation import sanitize_in_nd_realfloating

__all__ = ["rsvd", "reigh"]


##############################################################################################
# Randomized SVD "rSVD"
##############################################################################################


def _randomized_range_finder(
    A: DNDarray,
    rank: int,
    n_oversamples: int = 10,
    power_iter: int = 0,
    qr_procs_to_merge: int = 2,
) -> DNDarray:
    """
    Randomized range finder with q power iterations, stabilized with QR decompositions.
    Computes an orthonormal matrix Q with r columns whose range approximates the range of A.
    n_oversamples is the number of additional samples used to improve the quality of the approximation.
    For qr_procs_to_merge see the corresponding remarks for :func:`heat.linalg.qr() <heat.core.linalg.qr.qr()>`.

    Please note that "rank" here refers to the mathematical rank of a matrix, which is different from the notion of "(MPI-)rank".
    """
    if not isinstance(rank, int):
        raise TypeError(f"rank must be an integer, but is {type(rank)}.")
    if rank < 1:
        raise ValueError(f"rank must be positive, but is {rank}.")
    if not isinstance(n_oversamples, int):
        raise TypeError(
            f"if provided, n_oversamples must be an integer, but is {type(n_oversamples)}."
        )
    if n_oversamples < 0:
        raise ValueError(f"n_oversamples must be non-negative, but is {n_oversamples}.")
    if not isinstance(power_iter, int):
        raise TypeError(f"if provided, power_iter must be an integer, but is {type(power_iter)}.")
    if power_iter < 0:
        raise ValueError(f"power_iter must be non-negative, but is {power_iter}.")

    sanitize_in_nd_realfloating(A, "A", [2])

    ell = rank + n_oversamples
    q = power_iter

    # random matrix
    splitOmega = 1 if A.split == 0 else 0
    Omega = randn(A.shape[1], ell, dtype=A.dtype, device=A.device, split=splitOmega)

    # compute the range of A
    Y = matmul(A, Omega)
    Q, _ = qr(Y, procs_to_merge=qr_procs_to_merge)

    # power iterations
    for _ in range(q):
        if Q.split is not None and Q.shape[Q.split] < Q.comm.size:
            Q.resplit_(None)
        Y = matmul(A.T, Q)
        Q, _ = qr(Y, procs_to_merge=qr_procs_to_merge)
        if Q.split is not None and Q.shape[Q.split] < Q.comm.size:
            Q.resplit_(None)
        Y = matmul(A, Q)
        Q, _ = qr(Y, procs_to_merge=qr_procs_to_merge)

    return Q


def rsvd(
    A: DNDarray,
    svd_rank: int,
    n_oversamples: int = 10,
    power_iter: int = 0,
    qr_procs_to_merge: int = 2,
) -> Union[Tuple[DNDarray, DNDarray, DNDarray], Tuple[DNDarray, DNDarray]]:
    r"""
    Randomized SVD (rSVD) with prescribed truncation rank `svd_rank`.
    If :math:`A = U \operatorname{diag}(S) V^T` is the true SVD of A, this routine computes an approximation for U[:,:svd_rank] (and S[:svd_rank], V.T[:,:svd_rank]).

    The accuracy of this approximation depends on the structure of A ("low-rank" is best) and appropriate choice of parameters.

    Parameters
    ----------
    A : DNDarray
        2D-array (float32/64) of which the rSVD has to be computed.
    svd_rank : int
        truncation rank of the SVD. (This parameter corresponds to `n_components` in scikit-learn's TruncatedSVD.)
    n_oversamples : int, optional
        number of oversamples. The default is 10.
    power_iter : int, optional
        number of power iterations. The default is 0.
        Choosing `power_iter > 0` can improve the accuracy of the SVD approximation in the case of slowly decaying singular values, but increases the computational cost.
    qr_procs_to_merge : int, optional
        number of processes to merge at each step of QR decomposition in the power iteration (if power_iter > 0). The default is 2. See the corresponding remarks for :func:`heat.linalg.qr() <heat.core.linalg.qr.qr()>` for more details.


    Notes
    -----
    Memory requirements: the SVD computation of a matrix of size (svd_rank + n_oversamples) x (svd_rank + n_oversamples) must fit into the memory of a single process.
    The implementation follows Algorithm 4.4 (randomized range finder) and Algorithm 5.1 (direct SVD) in [1].

    Please note that "rank" in the context of SVD always refers to the number of singular values/vectors to compute (i.e., "rank" refers to the mathematical rank
    of a matrix). This is completely different from the notion of "(MPI-)rank", i.e., the ID given to a process, in a parallel MPI-application.

    References
    ----------
    [1] Halko, N., Martinsson, P. G., & Tropp, J. A. (2011). Finding structure with randomness: Probabilistic algorithms for constructing approximate matrix decompositions. SIAM review, 53(2), 217-288.
    """
    Q = _randomized_range_finder(
        A,
        svd_rank,
        n_oversamples=n_oversamples,
        power_iter=power_iter,
        qr_procs_to_merge=qr_procs_to_merge,
    )

    # compute the SVD of the projected matrix
    if Q.split is not None and Q.shape[Q.split] < Q.comm.size:
        Q.resplit_(None)
    B = matmul(Q.T, A)
    B.resplit_(
        None
    )  # B will be of size ell x n and thus small enough to fit into memory of a single process
    U, sigma, Vt = svd(B)
    U = matmul(Q, U)[:, :svd_rank]
    U.balance_()
    S = sigma[:svd_rank]
    Vt = Vt[:svd_rank, :]
    Vt.balance_()
    return U, S, Vt


def reigh(
    A: DNDarray,
    n_eigenvalues: int,
    n_oversamples: int = 10,
    power_iter: int = 0,
    qr_procs_to_merge: int = 2,
) -> Tuple[DNDarray, DNDarray]:
    r"""
    Randomized eigenvalue decomposition (rEIGH). Only the top `n_eigenvalues` eigenvalues (ordered by magnitude) and corresponding eigenvectors are computed.

    Parameters
    ----------
    A : DNDarray
        2D-array (float32/64) of which the rEIGH has to be computed. Must be symmetric.
    n_eigenvalues : int
        number of eigenvalues to compute. (This parameter corresponds to `n_components` in scikit-learn's PCA.)
    n_oversamples : int, optional
        number of oversamples. The default is 10.
    power_iter : int, optional
        number of power iterations. The default is 0.
        Choosing `power_iter > 0` can improve the accuracy of the eigenvalue approximation in the case of slowly decaying eigenvalues, but increases the computational cost.
    qr_procs_to_merge : int, optional
        number of processes to merge at each step of QR decomposition in the power iteration (if power_iter > 0). The default is 2. See the corresponding remarks for :func:`heat.linalg.qr() <heat.core.linalg.qr.qr()>` for more details.


    Notes
    -----
    Memory requirements: the symmetric eigenvalue decomposition of a matrix of size (n_eigenvalues + n_oversamples) x (n_eigenvalues + n_oversamples) must fit into the memory of a single process.
    The implementation follows Algorithm 4.4 (randomized range finder) and Algorithm 5.3 (eigenvalue decomposition from an SVD) in [1].

    References
    ----------
    [1] Halko, N., Martinsson, P. G., & Tropp, J. A. (2011). Finding structure with randomness: Probabilistic algorithms for constructing approximate matrix decompositions.
    SIAM review, 53(2), 217-288.
    """
    Q = _randomized_range_finder(
        A,
        n_eigenvalues,
        n_oversamples=n_oversamples,
        power_iter=power_iter,
        qr_procs_to_merge=qr_procs_to_merge,
    )
    # compute the eigenvalue decomposition of the projected matrix
    if Q.split is not None and Q.shape[Q.split] < Q.comm.size:
        Q.resplit_(None)
    B = matmul(Q.T, matmul(A, Q))
    B.resplit_(
        None
    )  # B will be of size ell x ell and thus small enough to fit into memory of a single process
    eigvals, eigvecs = torch.linalg.eigh(
        B.larray
    )  # actually just torch eigh as input is not split anymore
    idx = torch.argsort(eigvals, descending=True)
    eigvals = eigvals[idx][:n_eigenvalues]
    eigvecs = eigvecs[:, idx][:, :n_eigenvalues]
    eigvecs = DNDarray(
        eigvecs,
        tuple(eigvecs.shape),
        dtype=A.dtype,
        split=None,
        device=A.device,
        comm=A.comm,
        balanced=A.balanced,
    )
    eigvals = DNDarray(
        eigvals,
        tuple(eigvals.shape),
        dtype=A.dtype,
        split=None,
        device=A.device,
        comm=A.comm,
        balanced=A.balanced,
    )
    eigvecs = matmul(Q, eigvecs)
    return eigvals, eigvecs



import heat as ht

bool = ht.bool
"""Boolean (``True`` or ``False``)."""
int8 = ht.int8
"""An 8-bit signed integer whose values exist on the interval ``[-128, +127]``."""
int16 = ht.int16
"""A 16-bit signed integer whose values exist on the interval ``[-32,767, +32,767]``."""
int32 = ht.int32
"""A 32-bit signed integer whose values exist on the interval ``[-2,147,483,647, +2,147,483,647]``."""
int64 = ht.int64
"""A 64-bit signed integer whose values exist on the interval ``[-9,223,372,036,854,775,807, +9,223,372,036,854,775,807]``."""
uint8 = ht.uint8
"""An 8-bit unsigned integer whose values exist on the interval ``[0, +255]``."""
# For the status of Pytorch support for these 3 data types,
# see https://github.com/pytorch/pytorch/issues/58734
# uint16 =
# """A 16-bit unsigned integer whose values exist on the interval ``[0, +65,535]``."""
# uint32 =
# """A 32-bit unsigned integer whose values exist on the interval ``[0, +4,294,967,295]``."""
# uint64 =
# """A 64-bit unsigned integer whose values exist on the interval ``[0, +18,446,744,073,709,551,615]``."""
float32 = ht.float32
"""IEEE 754 single-precision (32-bit) binary floating-point number (see IEEE 754-2019)."""
float64 = ht.float64
"""IEEE 754 double-precision (64-bit) binary floating-point number (see IEEE 754-2019)."""
complex64 = ht.complex64
"""Single-precision (64-bit) complex floating-point number whose real and imaginary components must be IEEE 754 single-precision (32-bit) binary floating-point numbers (see IEEE 754-2019)."""
complex128 = ht.complex128
"""Double-precision (128-bit) complex floating-point number whose real and imaginary components must be IEEE 754 double-precision (64-bit) binary floating-point numbers (see IEEE 754-2019)."""

default_int = int64
"""Default integer data type is ``int64``"""
default_float = float64
"""Default floating-point data type is ``float64``"""

_all_dtypes = (
    bool,
    int8,
    int16,
    int32,
    int64,
    uint8,
    # uint16,
    # uint32,
    # uint64,
    float32,
    float64,
    complex64,
    complex128,
)
_boolean_dtypes = (bool,)
_real_floating_dtypes = (float32, float64)
_floating_dtypes = (float32, float64, complex64, complex128)
_complex_floating_dtypes = (complex64, complex128)
_integer_dtypes = (
    int8,
    int16,
    int32,
    int64,
    uint8,
    # uint16,
    # uint32,
    # uint64
)
_signed_integer_dtypes = (int8, int16, int32, int64)
_unsigned_integer_dtypes = (
    uint8,
    # uint16,
    # utnt32,
    # uint64
)
_integer_or_boolean_dtypes = (
    bool,
    int8,
    int16,
    int32,
    int64,
    uint8,
    # uint16,
    # uint32,
    # uint64,
)
_real_numeric_dtypes = (
    float32,
    float64,
    int8,
    int16,
    int32,
    int64,
    uint8,
    # uint16,
    # uint32,
    # uint64,
)
_numeric_dtypes = (
    float32,
    float64,
    complex64,
    complex128,
    int8,
    int16,
    int32,
    int64,
    uint8,
    # uint16,
    # uint32,
    # uint64,
)

_dtype_categories = {
    "all": _all_dtypes,
    "real numeric": _real_numeric_dtypes,
    "numeric": _numeric_dtypes,
    "integer": _integer_dtypes,
    "integer or boolean": _integer_or_boolean_dtypes,
    "boolean": _boolean_dtypes,
    "real floating-point": _floating_dtypes,
    "complex floating-point": _complex_floating_dtypes,
    "floating-point": _floating_dtypes,
}

_promotion_table = {
    (int8, int8): int8,
    (int8, int16): int16,
    (int8, int32): int32,
    (int8, int64): int64,
    (int16, int8): int16,
    (int16, int16): int16,
    (int16, int32): int32,
    (int16, int64): int64,
    (int32, int8): int32,
    (int32, int16): int32,
    (int32, int32): int32,
    (int32, int64): int64,
    (int64, int8): int64,
    (int64, int16): int64,
    (int64, int32): int64,
    (int64, int64): int64,
    (uint8, uint8): uint8,
    # (uint8, uint16): uint16,
    # (uint8, uint32): uint32,
    # (uint8, uint64): uint64,
    # (uint16, uint8): uint16,
    # (uint16, uint16): uint16,
    # (uint16, uint32): uint32,
    # (uint16, uint64): uint64,
    # (uint32, uint8): uint32,
    # (uint32, uint16): uint32,
    # (uint32, uint32): uint32,
    # (uint32, uint64): uint64,
    # (uint64, uint8): uint64,
    # (uint64, uint16): uint64,
    # (uint64, uint32): uint64,
    # (uint64, uint64): uint64,
    (int8, uint8): int16,
    # (int8, uint16): int32,
    # (int8, uint32): int64,
    (int16, uint8): int16,
    # (int16, uint16): int32,
    # (int16, uint32): int64,
    (int32, uint8): int32,
    # (int32, uint16): int32,
    # (int32, uint32): int64,
    (int64, uint8): int64,
    # (int64, uint16): int64,
    # (int64, uint32): int64,
    (uint8, int8): int16,
    # (uint16, int8): int32,
    # (uint32, int8): int64,
    (uint8, int16): int16,
    # (uint16, int16): int32,
    # (uint32, int16): int64,
    (uint8, int32): int32,
    # (uint16, int32): int32,
    # (uint32, int32): int64,
    (uint8, int64): int64,
    # (uint16, int64): int64,
    # (uint32, int64): int64,
    (float32, float32): float32,
    (float32, float64): float64,
    (float64, float32): float64,
    (float64, float64): float64,
    (complex64, complex64): complex64,
    (complex128, complex128): complex128,
    (complex64, complex128): complex128,
    (complex128, complex64): complex128,
    (float32, complex64): complex64,
    (float32, complex128): complex128,
    (float64, complex64): complex128,
    (float64, complex128): complex128,
    (complex64, float32): complex64,
    (complex64, float64): complex128,
    (complex128, float32): complex128,
    (complex128, float64): complex128,
    (bool, bool): bool,
}


def _result_type(type1, type2):
    if (type1, type2) in _promotion_table:
        return _promotion_table[type1, type2]
    raise TypeError(f"{type1} and {type2} cannot be type promoted together")



"""
Import all linear algebra functions into the ht.linalg namespace
"""

from .basics import *
from .solver import *
from .qr import *
from .svdtools import *
from .svd import *
from .polar import *
from .eigh import *
from .randomized import *
